#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## gcs.py
##
"""
    Interface to the Glasgow Constraint Solver's API for the CPMpy library.

    See:
    https://github.com/ciaranm/glasgow-constraint-solver

    The key feature of this CP solver is the ability to produce proof logs.

    Always use :func:`cp.SolverLookup.get("gcs") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'gcspy' python package is installed:

    .. code-block:: console

        $ pip install gcspy

    Source installation instructions:

    - Ensure you have C++20 compiler such as GCC 10.3  / clang 15
    - (on Debian-based systems, see https://apt.llvm.org for easy installation)
    - If necessary ``export CXX=<your up to date C++ compiler (e.g. clang++-15)>``
    - Ensure you have Boost installed
    - ``git clone https://github.com/ciaranm/glasgow-constraint-solver.git``
    - ``cd glasgow-constraint-solver/python``
    - ``pip install .``

    .. note::
        If for any reason you need to retry the build, ensure you remove glasgow-constraints-solver/generator before rebuilding.

    For the verifier functionality, the 'veripb' tool is also required.
    See https://gitlab.com/MIAOresearch/software/VeriPB#installation for installation instructions of veripb. 

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_gcs
"""
from typing import Optional
import pkg_resources

from cpmpy.transformations.comparison import only_numexpr_equality
from cpmpy.transformations.reification import reify_rewrite, only_bv_reifies
from ..exceptions import NotSupportedError, GCSVerificationException
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.variables import _BoolVarImpl, _IntVarImpl, _NumVarImpl, NegBoolView, boolvar
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.utils import is_num, argval, argvals
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.get_variables import get_variables
from ..transformations.flatten_model import flatten_constraint, flatten_objective, get_or_make_var
from ..transformations.safening import no_partial_functions

from ..transformations.normalize import toplevel_list

# For proof file handling and verifying
import sys
from os import path
from shutil import which
import subprocess

class CPM_gcs(SolverInterface):
    """
    Interface to Glasgow Constraint Solver's API.

    Creates the following attributes (see parent constructor for more):

    - ``gcs`` : the gcspy solver object
    - ``objective_var`` : optional: the variable used as objective
    - ``proof_location`` : location of the last proof produced by the solver
    - ``proof_name`` : name of the last proof (means <proof_name>.opb and <proof_name>.pbp will be present at the proof location)
    - ``veripb_return_code`` : return code from the last VeriPB check.
    - ``proof_check_timeout`` : whether the last VeriPB check timed out.

    Documentation of the solver's own Python API is sparse, but example usage can be found at:
    https://github.com/ciaranm/glasgow-constraint-solver/blob/main/python/python_test.py
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import gcspy
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e
        
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('gcspy').version
        except pkg_resources.DistributionNotFound:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: None (not supported)
        """
        if not self.supported():
            raise Exception("CPM_gcs: Install the python package 'gcspy' to use this solver interface.")

        import gcspy

        assert(subsolver is None) # unless you support subsolvers, see pysat or minizinc

        # initialise the native solver object
        self.gcs = gcspy.GCS()
        self.objective_var = None
        self.proof_location = None
        self.proof_name = None
        self.proof_check_timeout = True
        self.veripb_return_code = False

        # initialise everything else and post the constraints/objective
        super().__init__(name="Glasgow Constraint Solver", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.gcs
    
    def has_objective(self):
        return self.objective_var is not None
    
    def solve(self, time_limit=None, prove=False, proof_name=None, proof_location=".", 
              verify=False, verify_time_limit=None, veripb_args = [], display_verifier_output=True, **kwargs):
        """
            Run the Glasgow Constraint Solver, get just one (optimal) solution.

            Arguments:
                time_limit (float, optional):   maximum solve time in seconds.
                prove:                          whether to produce a VeriPB proof (.opb model file and .pbp proof file).
                proof_name:                     name for the the proof files.
                proof_location:                 location for the proof files (default to current working directory).
                verify:                         whether to verify the result of the solve run (overrides prove if prove is False)
                verify_time_limit:              time limit for verification (ignored if verify=False) 
                veripb_args:                    list of command line arguments to pass to veripb e.g. ``--trace --useColor`` (run ``veripb --help`` for a full list)
                display_verifier_output:        whether to print the output from VeriPB
                **kwargs:                       currently GCS does not support any additional keyword arguments.

            Returns: 
                whether a solution was found.
        """
        # ensure all user vars are known to solver
        self.solver_vars(list(self.user_vars))
        
        # If we're verifying we must be proving
        prove |= verify
        # Set default proof name to name of file containing __main__
        if prove and proof_name is None:
            if hasattr(sys.modules['__main__'], "__file__"):
                self.proof_name = path.splitext(path.basename(sys.modules['__main__'].__file__))[0]
            else:
                self.proof_name = "gcs_proof"
        else:
            self.proof_name = proof_name
        self.proof_location = proof_location
     
        # set time limit
        if time_limit is not None and time_limit <= 0:
            raise ValueError("Time limit must be positive")
                 
        # call the solver, with parameters    
        self.gcs_result = self.gcs.solve(
            all_solutions=self.has_objective(), 
            timeout=time_limit,
            callback=None,
            prove=prove,
            proof_name=self.proof_name,
            proof_location=proof_location,
            **kwargs)

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.gcs_result["solve_time"]

        # translate exit status
        if self.gcs_result['solutions'] != 0:
            if self.gcs_result['completed'] and self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif not self.gcs_result['completed']:
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if isinstance(cpm_var, _BoolVarImpl):
                    # Convert back to bool
                    cpm_var._value = bool(self.gcs.get_solution_value(sol_var, self.gcs_result['solutions']-1))
                else:
                    cpm_var._value = self.gcs.get_solution_value(sol_var, self.gcs_result['solutions']-1)

            # translate objective, for optimisation problems only
            if self.has_objective():
                self.objective_value_ = self.gcs.get_solution_value(self.solver_var(self.objective_var))

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None

        # Verify proof, if requested
        if verify:

            # set time limit
            if verify_time_limit is not None:
                if verify_time_limit <= 0:
                    raise ValueError("Time limit for verifying must be positive")

            self.verify(name=self.proof_name, location=proof_location, time_limit=verify_time_limit,
                        veripb_args=veripb_args, display_output=display_verifier_output)
            
            if self.veripb_return_code > 0:
                raise GCSVerificationException("Glasgow Constraint Solver: Proof failed to verify.")
            
        return has_sol

    def solveAll(self, time_limit=None, display=None, solution_limit=None, call_from_model=False, 
                 prove=False, proof_name=None, proof_location=".", verify=False, verify_time_limit=None, veripb_args = [], 
                 display_verifier_output=True, **kwargs):
        """
            Run the Glasgow Constraint Solver, and get a number of solutions, with optional solution callbacks. 

            Arguments:
                display:                        either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                                                default/None: nothing displayed
                solution_limit:                 stop after this many solutions (default: None)
                time_limit (float, optional):   maximum solve time in seconds (default: None)
                call_from_model:                whether the method is called from a CPMpy Model instance or not
                prove:                          whether to produce a VeriPB proof (.opb model file and .pbp proof file).
                proof_name:                     name for the the proof files.
                proof_location:                 location for the proof files (default to current working directory).
                verify:                         whether to verify the result of the solve run (overrides prove if prove is False)
                verify_time_limit:              time limit for verification (ignored if verify=False) 
                veripb_args:                    list of command line arguments to pass to veripb e.g. ``--trace --useColor`` (run ``veripb --help`` for a full list)
                display_verifier_output:        whether to print the output from VeriPB
                **kwargs:                       currently GCS does not support any additional keyword arguments.

            Returns: 
                number of solutions found
        """
        if self.has_objective():
            raise NotSupportedError("Glasgow Constraint Solver: does not support finding all optimal solutions.")
        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

         # If we're verifying we must be proving
        prove |= verify
        # Set default proof name to name of file containing __main__
        if prove and proof_name is None:
            if hasattr(sys.modules['__main__'], "__file__"):
                self.proof_name = path.splitext(path.basename(sys.modules['__main__'].__file__))[0]
            else:
                self.proof_name = "gcs_proof"
        self.proof_location = proof_location

        # Set display callback
        def display_callback(solution_map):
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if isinstance(cpm_var, _BoolVarImpl):
                    # Convert back to bool
                    cpm_var._value = bool(solution_map[sol_var])
                else:
                    cpm_var._value = solution_map[sol_var]

            if isinstance(display, Expression):
                print(argval(display))
            elif isinstance(display, list):
                # explicit list of expressions to display
                print(argvals(display))
            elif callable(display):
                display()
            else:
                raise NotImplementedError("Glasgow Constraint Solver: Unknown display type.".format(cpm_var))
            return 
        sol_callback = None
        if display:
            sol_callback=display_callback

        self.gcs_result = self.gcs.solve(
            all_solutions=True, 
            timeout=time_limit, 
            solution_limit=solution_limit, 
            callback=sol_callback, 
            prove=prove, 
            proof_name=proof_name, 
            proof_location=proof_location, **kwargs)

        # new status, get runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.gcs_result["solve_time"]

        num_sols = self.gcs_result["solutions"]
        if self.gcs_result["completed"] and num_sols >= 1:
            self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        elif self.gcs_result["completed"] and num_sols == 0:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif num_sols >= 1:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        else: # maybe unsat, maybe not (maybe a timeout)
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN

        # clear user vars if no solution found
        if self._solve_return(self.cpm_status, self.objective_value_) is False:
            for var in self.user_vars:
                var._value = None

        # Verify proof, if requested
        if verify:
            self.verify(name=self.proof_name, location=proof_location, time_limit=verify_time_limit, 
                        veripb_args=veripb_args, display_output=display_verifier_output)

        return num_sols

    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var): # shortcut, eases posting constraints
            return self.gcs.create_integer_constant(cpm_var)

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            # gcs only works with integer variables, so not(x) = -x + 1
            return self.gcs.add_constant(self.gcs.negate(self.solver_var(cpm_var._bv)), 1)

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                # Bool vars are just int vars with [0, 1] domain
                revar = self.gcs.create_integer_variable(0, 1, str(cpm_var))
            elif isinstance(cpm_var, _IntVarImpl):
                revar = self.gcs.create_integer_variable(cpm_var.lb, cpm_var.ub, str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        return self._varmap[cpm_var]

    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize.

            ``objective()`` can be called multiple times, only the last one is stored.

            .. note::
                technical side note: any constraints created during conversion of the objective
                are permanently posted to the solver
        """
        # make objective function non-nested
        (flat_obj, flat_cons) = flatten_objective(expr, csemap=self._csemap)
        self += flat_cons # add potentially created constraints
        self.user_vars.update(get_variables(flat_obj)) # add objvars to vars

        (obj, obj_cons) = get_or_make_var(flat_obj, csemap=self._csemap)
        self += obj_cons

        self.objective_var = obj

        if minimize:
            self.gcs.minimise(self.solver_var(obj))  
        else:
            self.gcs.maximise(self.solver_var(obj))

    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        cpm_cons = toplevel_list(cpm_expr)
        supported = {
            "min", 
            "max", 
            "abs", 
            "alldifferent", 
            "element", 
            'table', 
            'negative_table', 
            'count', 
            'nvalue',
            'inverse', 
            'circuit', 
            'xor'}
        cpm_cons = no_partial_functions(cpm_cons)
        cpm_cons = decompose_in_tree(cpm_cons, supported, csemap=self._csemap)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form

        # NB: GCS supports full reification for linear equality and linear inequaltiy constraints
        # but no reification for linear not equals and not half reification for linear equality. 
        # Maybe a future transformation (or future work on the GCS solver).
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['==']), csemap=self._csemap)
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum"]), csemap=self._csemap)  # supports >, <, !=

        # NB: GCS supports a small number of simple expressions as the reifying term
        # e.g. (x > 3) -> constraint could in principle be supported in the future.
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        str_rep = ""
        for c in cpm_cons:
            str_rep += str(c) + '\n'
        return cpm_cons

    def verify(self, name=None, location=".", time_limit=None, display_output=False, veripb_args=[]):
        """
        Verify a solver-produced proof using VeriPB.

        Requires that the 'veripb' tool is installed and on system path. 
        See https://gitlab.com/MIAOresearch/software/VeriPB#installation for installation instructions. 

        Arguments:
            - name:             name for the the proof files (default to self.proof_name)
            - location:         location for the proof files (default to current working directory).
            - time_limit:       time limit for verification (ignored if verify=False) 
            - veripb_args:      list of command line arguments to pass to veripb e.g. ``--trace --useColor`` (run ``veripb --help`` for a full list)
            - display_output:   whether to print the output from VeriPB
        """
        if not which("veripb"):
            raise Exception("Unable to run VeriPB: check it is installed and on system path - see https://gitlab.com/MIAOresearch/software/VeriPB#installation.")

        if name is None:
            name = self.proof_name
        
        if name is None: # Still None?
            raise ValueError("No proof to verify")
        
        if not isinstance(veripb_args, list):
            raise ValueError("veripb_args should be a list")
        
        opb_file = path.join(location, name +".opb")
        pbp_file = path.join(location, name +".pbp")

        if not path.isfile(opb_file):
            raise FileNotFoundError("Can't find " + opb_file)
        if not path.isfile(pbp_file):
            raise FileNotFoundError("Can't find " + pbp_file)
        
        try:
            result = subprocess.run(["veripb"] + veripb_args + [opb_file, pbp_file], 
                                    capture_output=True, text=True, timeout=time_limit)
            self.proof_check_timeout = False
            self.veripb_return_code = result.returncode
            if display_output:
                print(result.stdout)
                print(result.stderr)
        except subprocess.TimeoutExpired:
            self.proof_check_timeout = True
            self.veripb_return_code = 0

        return self.veripb_return_code
    
    def add(self, cpm_cons):
        """
        Post a (list of) CPMpy constraints(=expressions) to the solver
        Note that we don't store the constraints in a cpm_model,
        we first transform the constraints into primitive constraints,
        then post those primitive constraints directly to the native solver
        :param cpm_con CPMpy constraint, or list thereof
        :type cpm_con (list of) Expression(s)
        """
        # add new user vars to the set
                # add new user vars to the set
        get_variables(cpm_cons, collect=self.user_vars)

        for con in self.transform(cpm_cons):
            cpm_expr = con
            if isinstance(cpm_expr, _BoolVarImpl):
                # base case, just var or ~var
                self.gcs.post_or([self.solver_var(cpm_expr)])
            elif isinstance(cpm_expr, BoolVal):
                if not cpm_expr:
                    # bit a hack, empty clause does not work (issue #73 on gcs github)
                    a = boolvar()
                    self.gcs.post_and(self.solver_vars([a,~a]))
            elif isinstance(cpm_expr, Operator) or \
                (cpm_expr.name == '==' and isinstance(cpm_expr.args[0], _BoolVarImpl) \
                and not isinstance(cpm_expr.args[1], _NumVarImpl)): 
                # ^ Somewhat awkward, but want to deal with full and half reifications
                # in one go here, and then deal with regular == comparisons later.l

                # 'and'/n, 'or'/n, '->'/2
                if cpm_expr.name == 'and':
                    self.gcs.post_and(self.solver_vars(cpm_expr.args))
                elif cpm_expr.name == 'or':
                    self.gcs.post_or(self.solver_vars(cpm_expr.args))

                # Reified constraint: BoolVar -> Boolexpr or BoolVar == Boolexpr
                # LHS must be boolvar due to only_bv_reifies
                elif cpm_expr.name == '->' or cpm_expr.name == '==':
                    fully_reify = (cpm_expr.name == '==')
                    assert(isinstance(cpm_expr.args[0], _BoolVarImpl))  
                    bool_lhs = cpm_expr.args[0]
                    reif_var = self.solver_var(bool_lhs)
                    bool_expr = cpm_expr.args[1]

                    # Just a plain implies or equals between two boolvars
                    if isinstance(bool_expr, _BoolVarImpl): # bv1 -> bv2
                        (self.gcs.post_equals if fully_reify else self.gcs.post_implies)\
                            (*self.solver_vars([bool_lhs, bool_expr]))
                    elif isinstance(bool_expr, Operator): # bv -> and(...), bv -> or(...)  bv -> [->]
                        if bool_expr.name == 'and':
                            self.gcs.post_and_reif(self.solver_vars(bool_expr.args), reif_var, fully_reify)
                        elif bool_expr.name == 'or':
                            self.gcs.post_or_reif(self.solver_vars(bool_expr.args), reif_var, fully_reify)
                        elif bool_expr.name == '->':
                            self.gcs.post_implies_reif(self.solver_vars(bool_expr.args), reif_var, fully_reify)
                        else:
                            # Shouldn't happen if reify_rewrite worked?
                            raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}' {}".format)
                    
                    # Reified Comparison
                    elif isinstance(bool_expr, Comparison):
                        lhs = bool_expr.args[0]
                        rhs = bool_expr.args[1]
                        if bool_expr.name == '==':
                            self.gcs.post_equals_reif(*self.solver_vars([lhs, rhs]), reif_var, fully_reify)
                        elif bool_expr.name == '<=':
                            self.gcs.post_less_than_equal_reif(*self.solver_vars([lhs, rhs]), reif_var, fully_reify)
                        elif bool_expr.name == '<':
                            self.gcs.post_less_than_reif(*self.solver_vars([lhs, rhs]), reif_var, fully_reify)
                        elif bool_expr.name == '>=':
                            self.gcs.post_greater_than_equal_reif(*self.solver_vars([lhs, rhs]), reif_var, fully_reify)
                        elif bool_expr.name == '>':
                            self.gcs.post_greater_than_reif(*self.solver_vars([lhs, rhs]), reif_var, fully_reify)
                        elif bool_expr.name == '!=':
                            # Note: GCS doesn't currently support reified NotEquals, so we need this ugly workaround for now:
                            # bv -> x != y can be written as 
                            # bv -> OR(lt, gt) with lt, gt being BoolVars and the additional constraints
                            # lt == x < y
                            # gt == x > y
                            lt_bool, gt_bool = boolvar(shape=2)
                            self += (lhs < rhs) == lt_bool
                            self += (lhs > rhs) == gt_bool
                            if fully_reify:
                                self += (~bool_lhs).implies(lhs == rhs)
                            self.gcs.post_or_reif(self.solver_vars([lt_bool, gt_bool]), reif_var, False)
                        else:
                            raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}' {}".format)
                    else:
                        # Shouldn't happen if reify_rewrite worked
                        raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}' {}".format)
            
            # Normal comparison     
            elif isinstance(cpm_expr, Comparison):
                lhs = cpm_expr.args[0]
                rhs = cpm_expr.args[1]

                # Due to only_numexpr_equality we can only have '!=', "<=", etc.
                # when the lhs is a variable, sum or wsum
                if isinstance(lhs, _NumVarImpl) or lhs.name == 'sum' or lhs.name == 'wsum':
                    if lhs.name == 'sum' or lhs.name == 'wsum':
                        if lhs.name == 'sum':
                            summands = self.solver_vars(lhs.args)
                            summands.append(self.solver_var(rhs))
                            coeffs = [1]*len(lhs.args) + [-1]
                        else:
                            summands = self.solver_vars(lhs.args[1])
                            summands.append(self.solver_var(rhs))
                            coeffs = list(lhs.args[0]) + [-1]

                        if cpm_expr.name == '==':
                            self.gcs.post_linear_equality(summands, coeffs, 0)
                        elif cpm_expr.name == '!=':
                            self.gcs.post_linear_not_equal(summands, coeffs, 0)
                        elif cpm_expr.name == '<=':
                            self.gcs.post_linear_less_equal(summands, coeffs, 0)
                        elif cpm_expr.name == '<':
                            self.gcs.post_linear_less_equal(summands, coeffs, -1)
                        elif cpm_expr.name == '>=':
                            self.gcs.post_linear_greater_equal(summands, coeffs, 0)
                        elif cpm_expr.name == '>':
                            self.gcs.post_linear_greater_equal(summands, coeffs, 1)
                        else:
                            raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}'".format(cpm_expr))
                    else:
                        if cpm_expr.name == '==':
                            self.gcs.post_equals(*self.solver_vars([lhs, rhs]))
                        elif cpm_expr.name == '!=':
                            self.gcs.post_not_equals(*self.solver_vars([lhs, rhs]))
                        elif cpm_expr.name == '<=':
                            self.gcs.post_less_than_equal(*self.solver_vars([lhs, rhs]))
                        elif cpm_expr.name == '<':
                            self.gcs.post_less_than(*self.solver_vars([lhs, rhs]))
                        elif cpm_expr.name == '>=':
                            self.gcs.post_greater_than_equal(*self.solver_vars([lhs, rhs]))
                        elif cpm_expr.name == '>':
                            self.gcs.post_greater_than(*self.solver_vars([lhs, rhs]))
                        else:
                            raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}'".format(cpm_expr))

                # If the comparison is '==' we can have a NumExpr on the lhs
                elif cpm_expr.name == '==':
                    if lhs.name == 'abs':
                        assert(len(lhs.args) == 1) # Should not have a nested expression inside abs
                        self.gcs.post_abs(*self.solver_vars(list(lhs.args) + [rhs]))
                    elif lhs.name in ['mul', 'div', 'pow', 'mod']:
                        self.gcs.post_arithmetic(*self.solver_vars(list(lhs.args) + [rhs]), lhs.name)
                    elif lhs.name == 'sub':
                        var1 = self.solver_var(lhs.args[0])
                        nVar2 = self.gcs.negate(self.solver_var(lhs.args[1]))
                        self.gcs.post_arithmetic(var1, nVar2, self.solver_var(rhs), 'sum')
                    elif lhs.name == 'sum' and len(lhs.args) == 2:
                        var1 = self.solver_var(lhs.args[0])
                        var2 = self.solver_var(lhs.args[1])
                        self.gcs.post_arithmetic(var1, var2, self.solver_var(rhs), 'sum')
                    elif lhs.name == 'sum' and len(lhs.args) > 2:
                        summands = self.solver_vars(lhs.args)
                        summands.append(self.gcs.negate(self.solver_var(rhs)))
                        self.gcs.post_linear_equality(summands, [1]*len(summands), 0)
                    elif lhs.name == 'wsum':
                        summands = self.solver_vars(lhs.args[1])
                        summands.append(self.gcs.negate(self.solver_var(rhs)))
                        self.gcs.post_linear_equality(summands, list(lhs.args[0]) + [1], 0)
                    elif lhs.name == 'max':
                        self.gcs.post_max(self.solver_vars(lhs.args), self.solver_var(rhs))
                    elif lhs.name == 'min':
                        self.gcs.post_min(self.solver_vars(lhs.args), self.solver_var(rhs))   
                    elif lhs.name == 'element':
                        self.gcs.post_element(self.solver_var(rhs), self.solver_vars(lhs.args[1]), self.solver_vars(lhs.args[0])) 
                    elif lhs.name == 'count':
                        self.gcs.post_count(self.solver_vars(lhs.args[0]), self.solver_var(lhs.args[1]), self.solver_var(rhs))
                    elif lhs.name == 'nvalue':
                        self.gcs.post_nvalue(self.solver_var(rhs), self.solver_vars(lhs.args))
                    else:
                        # Think that's all the possible NumExprs?
                        raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}'".format(cpm_expr))
                else:
                    raise NotImplementedError("Not currently supported by Glasgow Constraint Solver API '{}'".format(cpm_expr))
            
            # rest: base (Boolean) global constraints
            elif cpm_expr.name == 'xor':
                self.gcs.post_xor(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == 'circuit':
                self.gcs.post_circuit(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == 'inverse':
                self.gcs.post_inverse(self.solver_vars(cpm_expr.args[0]), self.solver_vars(cpm_expr.args[1]))
            elif cpm_expr.name == 'alldifferent':
                self.gcs.post_alldifferent(self.solver_vars(cpm_expr.args))
            elif cpm_expr.name == 'table':
                self.gcs.post_table(self.solver_vars(cpm_expr.args[0]), cpm_expr.args[1])
            elif cpm_expr.name == 'negative_table':
                self.gcs.post_negative_table(self.solver_vars(cpm_expr.args[0]), cpm_expr.args[1])
            elif isinstance(cpm_expr, GlobalConstraint):
                # GCS also has SmartTable, Regular Language Membership, Knapsack constraints
                # which could be added in future. 
                self += cpm_expr.decompose()  # assumes a decomposition exists...
            else:
                # Hopefully we don't end up here.
                raise NotImplementedError(cpm_expr)

        return self
    __add__ = add  # avoid redirect in superclass


        



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## z3.py
##
"""
    Interface to Z3's Python API.

    Z3 is a highly versatile and effective theorem prover from Microsoft.
    Underneath, it is an SMT solver with a wide scala of theory solvers.
    We will interface to the finite-domain integer related parts of the API.
    (see https://github.com/Z3Prover/z3)

    .. warning::
        For incrementally solving an optimisation function, instantiate the solver object
        with a model that has an objective function, e.g. ``s = cp.SolverLookup.get("z3", Model(maximize=1))``.

    Always use :func:`cp.SolverLookup.get("z3") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'z3-solver' python package is installed:

    .. code-block:: console
    
        $ pip install z3-solver

    See detailed installation instructions at:
    https://github.com/Z3Prover/z3#python

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_z3

    ==============
    Module details
    ==============
"""
from typing import Optional
import pkg_resources

from cpmpy.transformations.get_variables import get_variables
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.globalconstraints import GlobalConstraint, DirectConstraint
from ..expressions.globalfunctions import GlobalFunction
from ..expressions.variables import _BoolVarImpl, NegBoolView, _NumVarImpl, _IntVarImpl, intvar
from ..expressions.utils import is_num, is_any_list, is_bool, is_int, is_boolexpr, eval_comparison
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.normalize import toplevel_list
from ..transformations.safening import no_partial_functions


class CPM_z3(SolverInterface):
    """
    Interface to Z3's Python API.

    Creates the following attributes (see parent constructor for more):
        
    - ``z3_solver``: object, z3's Solver() object

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, calls a function in the `z3` namespace and ``z3_solver.add()``'s the result.

    Documentation of the solver's own Python API:
    https://z3prover.github.io/api/html/namespacez3py.html

    .. note::
        Terminology note: a 'model' for z3 is a solution!
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import z3
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e
        
    @classmethod
    def version(cls) -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('z3-solver').version
        except pkg_resources.DistributionNotFound:
            return None
        
    def __init__(self, cpm_model=None, subsolver="sat"):
        """
        Constructor of the native solver object

        Arguments:
        - cpm_model: Model(), a CPMpy Model() (optional)
        - subsolver: None
        """
        if not self.supported():
            raise Exception("CPM_z3: Install the python package 'z3-solver' to use this solver interface.")

        import z3

        if subsolver is None:
            if cpm_model and cpm_model.has_objective():
                subsolver = "opt"
            else:
                subsolver = "sat"
        assert "sat" in subsolver or "opt" in subsolver, "Z3 only has a satisfaction or optimization sub-solver."

        # initialise the native solver object
        if "sat" in subsolver:
            self.z3_solver = z3.Solver()
        if "opt" in subsolver:
            self.z3_solver = z3.Optimize()

        # handle of objective (as returned by solver)
        self.obj_handle = None

        # initialise everything else and post the constraints/objective
        super().__init__(name="z3", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.z3_solver


    def solve(self, time_limit=None, assumptions=[], **kwargs):
        """
            Call the z3 solver

            Arguments:
                time_limit (float, optional):       maximum solve time in seconds
                assumptions:                        list of CPMpy Boolean variables (or their negation) that are assumed to be true.
                                                    For repeated solving, and/or for use with :func:`s.get_core() <get_core()>`: if the model is UNSAT,
                                                    get_core() returns a small subset of assumption variables that are unsat together.
                **kwargs:                           any keyword argument, sets parameters of solver object

            Arguments that correspond to solver parameters:

            - ... (no common examples yet)

            The full list doesn't seem to be documented online, you have to run its help() function:
            
            .. code-block:: python

                import z3
                z3.Solver().help()

            .. warning::
                Warning! Some parameternames in z3 have a '.' in their name,
                such as (arbitrarily chosen): ``sat.lookahead_simplify``
                You have to construct a dictionary of keyword arguments upfront:
                
                .. code-block:: python

                    params = {"sat.lookahead_simplify": True}
                    s.solve(**params)
        """
        import z3

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            # z3 expects milliseconds in int
            self.z3_solver.set(timeout=int(time_limit*1000))


        z3_assum_vars = self.solver_vars(assumptions)
        self.assumption_dict = {z3_var : cpm_var for (cpm_var, z3_var) in zip(assumptions, z3_assum_vars)}


        # call the solver, with parameters
        for (key,value) in kwargs.items():
            self.z3_solver.set(key, value)

        my_status = repr(self.z3_solver.check(*z3_assum_vars))

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        st = self.z3_solver.statistics()
        if 'time' not in st.keys():
            self.cpm_status.runtime = 0
        else:
            self.cpm_status.runtime = st.get_key_value('time')

        # translate exit status
        if my_status == "sat":
            if self.has_objective(): # COP
                # check if optimal solution found and proven, i.e. bounds are equal
                lower_bound = self.z3_solver.lower(self.obj_handle)
                upper_bound = self.z3_solver.upper(self.obj_handle)
                if lower_bound == upper_bound: # found optimal
                    self.cpm_status.exitstatus = ExitStatus.OPTIMAL
                else: # suboptimal / not proven
                    self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            else: # CSP
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif my_status == "unsat":
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif my_status == "unknown":
            try:
                model = self.z3_solver.model()
                if model: # a solution was found, just not the optimal one (or not proven)
                    self.cpm_status.exitstatus = ExitStatus.FEASIBLE
                # can happen when timeout is reached...
                else:
                    self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            # can happen when timeout is reached...
            except z3.Z3Exception as e: # no model has been initialized, not even an empty one
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN

        else:  # another?
            raise NotImplementedError(my_status)  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            sol = self.z3_solver.model() # the solution (called model in z3)
            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if isinstance(cpm_var, _BoolVarImpl):
                    cpm_var._value = bool(sol[sol_var])
                elif isinstance(cpm_var, _NumVarImpl):
                    cpm_var._value = sol[sol_var].as_long()

            # translate objective, for optimisation problems only
            if self.has_objective():
                obj = self.z3_solver.objectives()[0]
                self.objective_value_ = sol.evaluate(obj).as_long()

        else:  # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        import z3

        if is_num(cpm_var): # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return z3.Not(self.solver_var(cpm_var._bv))

        # create if it does not exit
        if cpm_var not in self._varmap:
            # we assume al variables are user variables (because nested expressions)
            self.user_vars.add(cpm_var)
            if isinstance(cpm_var, _BoolVarImpl):
                revar = z3.Bool(str(cpm_var))
            elif isinstance(cpm_var, _IntVarImpl):
                revar = z3.Int(str(cpm_var))
                # set bounds
                self.z3_solver.add(revar >= cpm_var.lb)
                self.z3_solver.add(revar <= cpm_var.ub)
            else:
                raise NotImplementedError("Not a know var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        return self._varmap[cpm_var]


    def has_objective(self):
        import z3
        return isinstance(self.z3_solver, z3.Optimize) and len(self.z3_solver.objectives()) != 0

    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            ``objective()`` can be called multiple times, only the last one is stored

            .. note::
                technical side note: any constraints created during conversion of the objective
                are premanently posted to the solver
        """
        import z3
        # objective can be a nested expression for z3
        if not isinstance(self.z3_solver, z3.Optimize):
            raise NotSupportedError("Use the z3 optimizer for optimization problems")

        if isinstance(expr, GlobalFunction): # not supported by Z3
            obj_var = intvar(*expr.get_bounds())
            self += expr == obj_var
            expr = obj_var

        obj = self._z3_expr(expr)
        if minimize:
            self.obj_handle = self.z3_solver.minimize(obj)
        else:
            self.obj_handle = self.z3_solver.maximize(obj)


    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """

        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel={"div", "mod"})
        supported = {"alldifferent", "xor", "ite"}  # z3 accepts these reified too
        cpm_cons = decompose_in_tree(cpm_cons, supported, supported, csemap=self._csemap)
        return cpm_cons

    def add(self, cpm_expr):
        """
            Z3 supports nested expressions so translate expression tree and post to solver API directly

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
        """
        # all variables are user variables, handled in `solver_var()`
        # unless their constraint gets simplified away, so lets collect them anyway
        get_variables(cpm_expr, collect=self.user_vars)

        # transform and post the constraints
        for cpm_con in self.transform(cpm_expr):
            # translate each expression tree, then post straight away
            z3_con = self._z3_expr(cpm_con)
            self.z3_solver.add(z3_con)

        return self
    __add__ = add  # avoid redirect in superclass

    def _z3_expr(self, cpm_con):
        """
            Z3 supports nested expressions,
            so we recursively translate our expressions to theirs.

            Accepts single constraints or a list thereof, return type changes accordingly.

        """
        import z3

        if is_num(cpm_con):
            # translate numpy to python native
            if is_bool(cpm_con):
                return bool(cpm_con)
            elif is_int(cpm_con):
                return z3.IntVal(int(cpm_con))
            return float(cpm_con)

        elif is_any_list(cpm_con):
            # arguments can be lists
            return [self._z3_expr(con) for con in cpm_con]

        elif isinstance(cpm_con, BoolVal):
            return cpm_con.args[0]

        elif isinstance(cpm_con, _NumVarImpl):
            return self.solver_var(cpm_con)

        # Operators: base (bool), lhs=numexpr, lhs|rhs=boolexpr (reified ->)
        elif isinstance(cpm_con, Operator):
            arity, _ = Operator.allowed[cpm_con.name]
            # 'and'/n, 'or'/n, '->'/2
            if cpm_con.name == 'and':
                return z3.And(self._z3_expr(cpm_con.args))
            elif cpm_con.name == 'or':
                return z3.Or(self._z3_expr(cpm_con.args))
            elif cpm_con.name == '->':
                return z3.Implies(*self._z3_expr(cpm_con.args))
            elif cpm_con.name == 'not':
                return z3.Not(self._z3_expr(cpm_con.args[0]))

            # 'sum'/n, 'wsum'/2
            elif cpm_con.name == 'sum':
                return z3.Sum(self._z3_expr(cpm_con.args))
            elif cpm_con.name == 'wsum':
                w = cpm_con.args[0]
                x = self._z3_expr(cpm_con.args[1])
                return z3.Sum([wi*xi for wi,xi in zip(w,x)])

            # 'sub'/2, 'mul'/2, 'div'/2, 'pow'/2, 'm2od'/2
            elif arity == 2 or cpm_con.name == "mul":
                assert len(cpm_con.args) == 2, "Currently only support multiplication with 2 vars"
                x, y = self._z3_expr(cpm_con.args)
                if isinstance(x, z3.BoolRef):
                    x = z3.If(x, 1, 0)
                if isinstance(y, z3.BoolRef):
                    y = z3.If(y, 1, 0)

                if cpm_con.name == 'sub':
                    return x - y
                elif cpm_con.name == "mul":
                    return x * y
                elif cpm_con.name == "div":
                    # z3 rounds towards negative infinity, need this hack when result is negative
                    return z3.If(z3.And(x >= 0, y >= 0), x / y,
                           z3.If(z3.And(x <= 0, y <= 0), -x / -y,
                           z3.If(z3.And(x >= 0, y <= 0), -(x / -y),
                           z3.If(z3.And(x <= 0, y >= 0), -(-x / y), 0))))

                elif cpm_con.name == "pow":
                    if not is_num(cpm_con.args[1]):
                        # tricky in Z3 not all power constraints are decidable
                        # solver will return 'unknown', even if theory is satisfiable.
                        # https://stackoverflow.com/questions/70289335/power-and-logarithm-in-z3
                        # raise error to be consistent with other solvers
                        raise NotSupportedError(f"Z3 only supports power constraint with constant exponent, got {cpm_con}")
                    return x ** y
                elif cpm_con.name == "mod":
                    # minimic modulo with integer division (round towards o)
                    return z3.If(z3.And(x >= 0), x % y,-(-x % y))

            # '-'/1
            elif cpm_con.name == "-":
                if is_boolexpr(cpm_con.args[0]):
                    return -z3.If(self._z3_expr(cpm_con.args[0]), 1, 0)
                return -self._z3_expr(cpm_con.args[0])

            else:
                raise NotImplementedError(f"Operator {cpm_con} not (yet) implemented for Z3, "
                                          f"please report on github if you need it")

        # Comparisons (just translate the subexpressions and re-post)
        elif isinstance(cpm_con, Comparison):
            lhs, rhs = cpm_con.args

            lhs_bexpr = is_boolexpr(lhs)
            rhs_bexpr = is_boolexpr(rhs)

            lhs, rhs = self._z3_expr(cpm_con.args)

            if cpm_con.name == "==" or cpm_con.name == "!=":
                # z3 supports bool <-> bool comparison but not bool <-> arith
                if lhs_bexpr and not rhs_bexpr:
                    # upcast lhs to integer
                    lhs = z3.If(lhs, 1, 0)
                elif rhs_bexpr and not lhs_bexpr:
                    # upcase rhs to integer
                    rhs = z3.If(rhs, 1, 0)
            else:
                # other comparisons are not supported on boolexpr
                if lhs_bexpr: # upcast lhs
                    lhs = z3.If(lhs, 1, 0)
                if rhs_bexpr: # upcase rhs
                    rhs = z3.If(rhs, 1, 0)

            # post the comparison
            return eval_comparison(cpm_con.name, lhs, rhs)

        # rest: base (Boolean) global constraints
        elif isinstance(cpm_con, GlobalConstraint):
            # TODO:
            # table

            if cpm_con.name == 'alldifferent':
                return z3.Distinct(self._z3_expr(cpm_con.args))
            elif cpm_con.name == 'xor':
                z3_args = self._z3_expr(cpm_con.args)
                if len(z3_args) == 1: # just the arg
                    return z3_args[0]
                z3_cons = z3.Xor(z3_args[0], z3_args[1])
                for a in z3_args[2:]:
                    z3_cons = z3.Xor(z3_cons, a)
                return z3_cons
            elif cpm_con.name == 'ite':
                return z3.If(self._z3_expr(cpm_con.args[0]), self._z3_expr(cpm_con.args[1]),
                             self._z3_expr(cpm_con.args[2]))

            raise ValueError(f"Global constraint {cpm_con} should be decomposed already, please report on github.")

        # a direct constraint, make with z3 (will be posted to it by calling function)
        elif isinstance(cpm_con, DirectConstraint):
            return cpm_con.callSolver(self, z3)

        raise NotImplementedError("Z3: constraint not (yet) supported", cpm_con)

    # Other functions from SolverInterface that you can overwrite:
    # solveAll, solution_hint, get_core

    def get_core(self):
        """
            For use with :func:`s.solve(assumptions=[...]) <solve()>`. Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

            CPMpy will return only those variables that are False (in the UNSAT core)

            Note that there is no guarantee that the core is minimal, though this interface does upon up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!
        """
        assert (self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE), "Can only extract core form UNSAT model"
        assert (len(self.assumption_dict) > 0), "Assumptions must be set using s.solve(assumptions=[...])"

        return [self.assumption_dict[z3_var] for z3_var in self.z3_solver.unsat_core()]






"""
Collection of callbacks for the PyCSP3 parser
"""

from functools import reduce

from pycsp3.parser.callbacks import Callbacks
from pycsp3.classes.auxiliary.conditions import Condition
from pycsp3.classes.auxiliary.enums import (TypeConditionOperator, TypeArithmeticOperator, TypeUnaryArithmeticOperator,
                                            TypeLogicalOperator,
                                            TypeOrderedOperator, TypeRank, TypeObj)
from pycsp3.classes.main.variables import Variable
from pycsp3.classes.nodes import (Node)

from pycsp3.parser.xentries import XVar
from pycsp3.tools.utilities import _Star

import cpmpy as cp
from cpmpy.tools.xcsp3 import globals as xglobals
from cpmpy import cpm_array
from cpmpy.expressions.utils import flatlist, get_bounds, is_boolexpr


class CallbacksCPMPy(Callbacks):  
    """
        A pycsp3-compatible callback for parsing XCSP3 instances into a CPMPy model.
    """

    def __init__(self):
        super().__init__()
        self.cpm_model = cp.Model()
        self.cpm_variables = dict()
        self.print_general_methods = False
        self.print_specific_methods = False

    def var_integer_range(self, x: Variable, min_value: int, max_value: int):
        if min_value == 0 and max_value == 1:
            # boolvar
            newvar = cp.boolvar(name=x.id)
        else:
            newvar = cp.intvar(min_value, max_value, name=x.id)
        self.cpm_variables[x] = newvar

    def var_integer(self, x: Variable, values: list[int]):
        mini = min(values)
        maxi = max(values)
        if mini == 0 and maxi == 1:
            # boolvar
            newvar = cp.boolvar(name=x.id)
        else:
            newvar = cp.intvar(mini, maxi, name=x.id)
        self.cpm_variables[x] = newvar
        nbvals = maxi - mini + 1
        if len(values) < nbvals:
            # only do this if there are holes in the domain
            self.cpm_model += cp.InDomain(newvar, values)  # faster decomp, only works in positive context

    def load_instance(self, discarded_classes=None):
        return self.cpm_model, self.cpm_variables

    def ctr_true(self, scope: list[Variable]):
        return cp.BoolVal(True)

    def ctr_false(self, scope: list[Variable]):
        assert False, "Problem as a constraint with 0 supports: " + str(scope)

    def ctr_intension(self, scope: list[Variable], tree: Node):
        cons = self.intentionfromtree(tree)
        self.cpm_model += cons

    funcmap = {
        # Arithmetic
        "neg": (1, lambda x: -x),
        "abs": (1, lambda x: abs(x)),
        "add": (0, lambda x: cp.sum(x)),
        "sub": (2, lambda x, y: x - y),
        "mul": (0, lambda x: reduce((lambda x, y: x * y), x)),
        "div": (2, lambda x, y: x // y),
        "mod": (2, lambda x, y: x % y),
        "sqr": (1, lambda x: x ** 2),
        "pow": (2, lambda x, y: x ** y),
        "min": (0, lambda x: cp.min(x)),
        "max": (0, lambda x: cp.max(x)),
        "dist": (2, lambda x, y: abs(x - y)),
        # Relational
        "lt": (2, lambda x, y: x < y),
        "le": (2, lambda x, y: x <= y),
        "ge": (2, lambda x, y: x >= y),
        "gt": (2, lambda x, y: x > y),
        "ne": (2, lambda x, y: x != y),
        "eq": (0, lambda x: x[0] == x[1] if len(x) == 2 else cp.AllEqual(x)),
        # Set
        'in': (2, lambda x, y: cp.InDomain(x, y)),  # could be mixed context here!
        'notin': (2, lambda x, y: xglobals.NotInDomain(x, y)),  # could be mixed context here!
        'set': (0, lambda x: list(set(x))),
        # TODO 'notin' is the only other set operator (negative indomain)
        # Logic
        "not": (1, lambda x: ~x),
        "and": (0, lambda x: cp.all(x)),
        "or": (0, lambda x: cp.any(x)),
        "xor": (0, lambda x: cp.Xor(x)),
        "iff": (0, lambda x: cp.all(a == b for a, b in zip(x[:-1], x[1:]))),
        "imp": (2, lambda x, y: x.implies(y)),
        # control
        "if": (3, lambda b, x, y: cp.IfThenElse(b, x, y) if is_boolexpr(x) and is_boolexpr(y)
        else xglobals.IfThenElseNum(b, x, y))
    }

    def eval_cpm_comp(self, lhs, op: TypeConditionOperator, rhs):
        if (op == TypeConditionOperator.IN) or (op == TypeConditionOperator.NOTIN):
            assert isinstance(rhs, list), f"Expected list as rhs but got {rhs}"

        arity, cpm_op = self.funcmap[op.name.lower()]
        if arity == 2:
            return cpm_op(lhs, rhs)
        elif arity == 0:
            return cpm_op([lhs, rhs])
        else:
            raise ValueError(f"Expected operator of arity 0 or 2 but got {cpm_op} which is of arity {arity}")

    def intentionfromtree(self, node):
        if isinstance(node, Node):
            if node.type.lowercase_name == 'var':
                return self.cpm_variables[node.cnt]
            if node.type.lowercase_name == 'int':
                return node.cnt
            arity, cpm_op = self.funcmap[node.type.lowercase_name]
            cpm_args = []
            for arg in node.cnt:
                cpm_args.append(self.intentionfromtree(arg))
            if arity != 0:
                return cpm_op(*cpm_args)
            return cpm_op(cpm_args)
        else:
            return node

    def ctr_primitive1a(self, x: Variable, op: TypeConditionOperator, k: int):
        assert op.is_rel()
        cpm_x = self.get_cpm_var(x)
        self.cpm_model += self.eval_cpm_comp(cpm_x, op, k)

    def ctr_primitive1b(self, x: Variable, op: TypeConditionOperator, term: list[int] | range):
        assert op.is_set()
        x = self.get_cpm_var(x)
        arity, cpm_op = self.funcmap[op.name.lower()]
        if isinstance(term, range):
            term = [x for x in term]  # list from range
        if arity == 2:
            self.cpm_model += cpm_op(x, term)
        elif arity == 0:
            self.cpm_model += cpm_op([x, term])
        else:
            self._unimplemented(x, op, term)

    def ctr_primitive1c(self, x: Variable, aop: TypeArithmeticOperator, p: int, op: TypeConditionOperator, k: int):
        assert op.is_rel()
        self.ctr_primitive3(x, aop, p, op, k)  # for cpmpy ints and vars are just interchangeable..

    def ctr_primitive2a(self, x: Variable, aop: TypeUnaryArithmeticOperator, y: Variable):
        # TODO this was unimplemented not sure if it should be aop(x) == y or x == aop(y)..
        arity, cpm_op = self.funcmap[aop.name.lower()]
        assert arity == 1, "unary operator expected"
        x = self.get_cpm_var(x)
        y = self.get_cpm_var(y)
        self.cpm_model += cpm_op(x) == y

    def ctr_primitive2b(self, x: Variable, aop: TypeArithmeticOperator, y: Variable, op: TypeConditionOperator, k: int):
        # (x aop y) rel k
        self.ctr_primitive3(x, aop, y, op, k)  # for cpmpy ints and vars are just interchangeable..

    def ctr_primitive2c(self, x: Variable, aop: TypeArithmeticOperator, p: int, op: TypeConditionOperator, y: Variable):
        # (x aop p) op y
        assert op.is_rel()
        self.ctr_primitive3(x, aop, p, op, y)  # for cpmpy ints and vars are just interchangeable..

    def ctr_primitive3(self, x: Variable, aop: TypeArithmeticOperator, y: Variable, op: TypeConditionOperator,
                       z: Variable):
        # (x aop y) op z
        assert op.is_rel()
        arity_op, cpm_op = self.funcmap[(op.name).lower()]
        arity, cpm_aop = self.funcmap[aop.name.lower()]
        x = self.get_cpm_var(x)
        y = self.get_cpm_var(y)
        z = self.get_cpm_var(z)
        if arity == 2:
            if arity_op == 2:
                self.cpm_model += cpm_op(cpm_aop(x, y), z)
            else:  # eq is arity 0, because of allequal global
                self.cpm_model += cpm_op([cpm_aop(x, y), z])
        elif arity == 0:
            if arity_op == 2:
                self.cpm_model += cpm_op(cpm_aop([x, y]), z)
            else:  # eq is arity 0, because of allequal global
                self.cpm_model += cpm_op([cpm_aop([x, y]), z])

    def ctr_logic(self, lop: TypeLogicalOperator, scope: list[Variable]):  # lop(scope)
        if lop == TypeLogicalOperator.AND:
            self.cpm_model += self.get_cpm_vars(scope)
        elif lop == TypeLogicalOperator.OR:
            self.cpm_model += cp.any(self.get_cpm_vars(scope))
        elif lop == TypeLogicalOperator.IFF:
            assert len(scope) == 2
            a, b = scope
            self.cpm_model += self.get_cpm_var(a) == self.get_cpm_var(b)
        elif lop == TypeLogicalOperator.IMP:
            assert len(scope) == 2
            a, b = scope
            self.cpm_model += self.get_cpm_var(a).implies(self.get_cpm_var(b))
        elif lop == TypeLogicalOperator.XOR:
            self.cpm_model += cp.Xor(self.get_cpm_vars(scope))
        else:
            self._unimplemented(lop, scope)

    def ctr_logic_reif(self, x: Variable, y: Variable, op: TypeConditionOperator, k: int | Variable):  # x = y <op> k
        assert op.is_rel()
        cpm_x, cpm_y, cpm_k = self.get_cpm_vars([x, y, k])
        self.cpm_model += self.eval_cpm_comp(cpm_x == cpm_y, op, cpm_k)

    def ctr_logic_eqne(self, x: Variable, op: TypeConditionOperator, lop: TypeLogicalOperator,
                       scope: list[Variable]):  # x = lop(scope) or x != lop(scope)
        cpm_x = self.get_cpm_var(x)
        arity, cpm_op = self.funcmap[(lop.name).lower()]
        if arity == 0:
            rhs = cpm_op(self.get_cpm_vars(scope))
        elif arity == 2:
            rhs = cpm_op(*self.get_cpm_vars(scope))
        else:
            self._unimplemented(lop, scope)

        self.cpm_model += self.eval_cpm_comp(cpm_x, op, rhs)

    def unroll(self, values):
        res = []
        for v in values:
            if isinstance(v, range):
                res.extend(v)
            else:
                res.append(v)
        return res

    def ctr_extension_unary(self, x: Variable, values: list[int], positive: bool, flags: set[str]):
        if len(values) == 1 and isinstance(values[0], range):
            values = list(eval(str(values[0])))

        if positive:
            # unary table constraint is just an inDomain
            if len(values) == 1:
                self.cpm_model += self.get_cpm_var(x) == values[0]
            else:
                self.cpm_model += cp.InDomain(self.get_cpm_var(x), self.unroll(values))  # faster decomp, only works in positive context
        else:
            # negative, so not in domain
            if len(values) == 1:
                self.cpm_model += self.get_cpm_var(x) != values[0]
            else:
                self.cpm_model += xglobals.NotInDomain(self.get_cpm_var(x), self.unroll(values))

    def ctr_extension(self, scope: list[Variable], tuples: list, positive: bool, flags: set[str]):
        def strwildcard(x):
            if isinstance(x, _Star):
                return '*'
            return x

        if 'starred' in flags:
            cpm_vars = self.vars_from_node(scope)
            exttuples = [tuple([strwildcard(x) for x in tup]) for tup in tuples]
            if positive:
                self.cpm_model += xglobals.RowSelectingShortTable(cpm_vars, exttuples)
            else:
                self.cpm_model += xglobals.NegativeShortTable(cpm_vars, exttuples)
        else:
            cpm_vars = self.vars_from_node(scope)
            if positive:
                self.cpm_model += xglobals.NonReifiedTable(cpm_vars, tuples)
            else:
                self.cpm_model += cp.NegativeTable(cpm_vars, tuples)

    def ctr_regular(self, scope: list[Variable], transitions: list, start_state: str, final_states: list[str]):
        self.cpm_model += xglobals.Regular(self.get_cpm_vars(scope), transitions, start_state, final_states)

    def ctr_mdd(self, scope: list[Variable], transitions: list):
        self.cpm_model += xglobals.MDD(self.get_cpm_vars(scope), transitions)

    def ctr_all_different(self, scope: list[Variable] | list[Node], excepting: None | list[int]):
        cpm_exprs = self.get_cpm_exprs(scope)
        if excepting is None:
            self.cpm_model += cp.AllDifferent(cpm_exprs)
        elif len(excepting) > 0:
            self.cpm_model += cp.AllDifferentExceptN(cpm_exprs, excepting)
        elif len(excepting) == 0:
            # just in case they get tricky
            self.cpm_model += cp.AllDifferent(cpm_exprs)
        else:  # unsupported for competition
            self._unimplemented(scope, excepting)

    def ctr_all_different_lists(self, lists: list[list[Variable]], excepting: None | list[list[int]]):
        if excepting is None:
            self.cpm_model += xglobals.AllDifferentLists([self.get_cpm_vars(lst) for lst in lists])
        else:
            self.cpm_model += xglobals.AllDifferentListsExceptN([self.get_cpm_vars(lst) for lst in lists], excepting)

    def ctr_all_different_matrix(self, matrix: list[list[Variable]], excepting: None | list[int]):
        import numpy as np
        for row in matrix:
            self.ctr_all_different(row, excepting)
        for col in np.array(matrix).T:
            self.ctr_all_different(col, excepting)

    def ctr_all_equal(self, scope: list[Variable] | list[Node], excepting: None | list[int]):
        if excepting is None:
            self.cpm_model += cp.AllEqual(self.get_cpm_exprs(scope))
        else:
            self.cpm_model += cp.AllEqualExceptN(self.get_cpm_exprs(scope), excepting)

    def ctr_ordered(self, lst: list[Variable], operator: TypeOrderedOperator,
                    lengths: None | list[int] | list[Variable]):

        cpm_vars = self.get_cpm_vars(lst)

        if lengths is None:
            if operator == TypeOrderedOperator.INCREASING:
                self.cpm_model += cp.Increasing(cpm_vars)
            elif operator == TypeOrderedOperator.STRICTLY_INCREASING:
                self.cpm_model += cp.IncreasingStrict(cpm_vars)
            elif operator == TypeOrderedOperator.DECREASING:
                self.cpm_model += cp.Decreasing(cpm_vars)
            elif operator == TypeOrderedOperator.STRICTLY_DECREASING:
                self.cpm_model += cp.DecreasingStrict(cpm_vars)
            else:
                self._unimplemented(lst, operator, lengths)

        # also handle the lengths parameter
        if lengths is not None:
            lengths = self.get_cpm_vars(lengths)
            if operator == TypeOrderedOperator.INCREASING:
                for x, l, y in zip(cpm_vars[:-1], lengths, cpm_vars[1:]):
                    self.cpm_model += x + l <= y
            elif operator == TypeOrderedOperator.STRICTLY_INCREASING:
                for x, l, y in zip(cpm_vars[:-1], lengths, cpm_vars[1:]):
                    self.cpm_model += x + l < y
            elif operator == TypeOrderedOperator.DECREASING:
                for x, l, y in zip(cpm_vars[:-1], lengths, cpm_vars[1:]):
                    self.cpm_model += x + l >= y
            elif operator == TypeOrderedOperator.STRICTLY_DECREASING:
                for x, l, y in zip(cpm_vars[:-1], lengths, cpm_vars[1:]):
                    self.cpm_model += x + l > y
            else:
                self._unimplemented(lst, operator, lengths)

    def ctr_lex_limit(self, lst: list[Variable], limit: list[int],
                      operator: TypeOrderedOperator):  # should soon enter XCSP3-core
        self.ctr_lex([lst, limit], operator)

    def ctr_lex(self, lists: list[list[Variable]], operator: TypeOrderedOperator):
        cpm_lists = [self.get_cpm_vars(lst) for lst in lists]
        if operator == TypeOrderedOperator.STRICTLY_INCREASING:
            self.cpm_model += cp.LexChainLess(cpm_lists)
        elif operator == TypeOrderedOperator.INCREASING:
            self.cpm_model += cp.LexChainLessEq(cpm_lists)
        elif operator == TypeOrderedOperator.STRICTLY_DECREASING:
            rev_lsts = list(reversed(cpm_lists))
            self.cpm_model += cp.LexChainLess(rev_lsts)
        elif operator == TypeOrderedOperator.DECREASING:
            rev_lsts = list(reversed(cpm_lists))
            self.cpm_model += cp.LexChainLessEq(rev_lsts)
        else:
            self._unimplemented(lists, operator)

    def ctr_lex_matrix(self, matrix: list[list[Variable]], operator: TypeOrderedOperator):
        import numpy as np
        # lex_chain on rows
        self.ctr_lex(matrix, operator)
        # lex chain on columns
        self.ctr_lex(np.array(matrix).T.tolist(), operator)

    def ctr_precedence(self, lst: list[Variable], values: None | list[int], covered: bool):
        if covered is True:  # not supported for competition
            self._unimplemented(lst, values, covered)

        cpm_vars = self.get_cpm_vars(lst)
        if values is None:  # assumed to be ordered set of all values collected from domains in lst
            lbs, ubs = get_bounds(cpm_vars)
            values = set()
            for lb, ub in zip(lbs, ubs):
                values.update(list(range(lb, ub + 1)))
            values = sorted(values)

        self.cpm_model += cp.Precedence(cpm_vars, values)

    def ctr_sum(self, lst: list[Variable] | list[Node], coefficients: None | list[int] | list[Variable],
                condition: Condition):
        # cpm_vars = []
        # if isinstance(lst[0], XVar):
        #     for xvar in lst:
        #         cpm_vars.append(self.cpm_variables[xvar])
        # else:
        #     cpm_vars = self.exprs_from_node(lst)
        # arity, op = self.funcmap[condition.operator.name.lower()]
        # if hasattr(condition, "variable"):
        #     rhs = condition.variable
        # elif hasattr(condition, 'value'):
        #     rhs = condition.value
        # elif hasattr(condition, 'max'):
        #     #operator = in
        #     rhs = [x for x in range(condition.min, condition.max + 1)]
        # else:
        #     pass
        # cpm_rhs = self.get_cpm_var(rhs)
        # if coefficients is None:
        #     cpsum = cp.sum(cpm_vars)
        # else:
        #     cp_coeffs = self.get_cpm_vars(coefficients)
        #     cpsum = cp.sum(cp.cpm_array(cpm_vars) * cp_coeffs)
        # if arity == 0:
        #     self.cpm_model += op([cpsum, cpm_rhs])
        # else:
        #     self.cpm_model += op(cpsum, cpm_rhs)

        import numpy as np
        if coefficients is None or len(coefficients) == 0:
            coefficients = np.ones(len(lst), dtype=int)  # TODO I guess, if wsums are preferred over sums
        elif isinstance(coefficients[0], Variable):  # convert to cpmpy var
            coefficients = cp.cpm_array(self.get_cpm_vars(coefficients))
        else:
            coefficients = np.array(coefficients)

        lhs = cp.sum(coefficients * self.get_cpm_exprs(lst))

        if (condition.operator == TypeConditionOperator.IN) or (condition.operator == TypeConditionOperator.NOTIN):
            from pycsp3.classes.auxiliary.conditions import ConditionInterval, ConditionSet
            assert isinstance(condition,
                              (ConditionInterval, ConditionSet)), "Competition only supports intervals when operator is `in` or `notin`"  # TODO and not in? (notin)
            if isinstance(condition, ConditionInterval):
                rhs = list(range(condition.min, condition.max + 1))
            else: # ConditionSet
                rhs = list(condition.t)
        else:
            rhs = self.get_cpm_var(condition.right_operand())

        self.cpm_model += self.eval_cpm_comp(lhs, condition.operator, rhs)

    def ctr_count(self, lst: list[Variable] | list[Node], values: list[int] | list[Variable], condition: Condition):
        # General case of count, can accept list of variables for any arg and any operator
        cpm_vars = self.get_cpm_exprs(lst)
        cpm_vals = self.get_cpm_vars(values)
        if condition.operator == TypeConditionOperator.IN or (condition.operator == TypeConditionOperator.NOTIN):
            from pycsp3.classes.auxiliary.conditions import ConditionInterval, ConditionSet
            assert isinstance(condition,
                              (ConditionInterval, ConditionSet)), "Competition only supports intervals when operator is `in` or `notin`"  # TODO and not in? (notin)
            if isinstance(condition, ConditionInterval):
                rhs = list(range(condition.min, condition.max + 1))
            else: # ConditionSet
                rhs = list(condition.t)
        else:
            rhs = self.get_cpm_var(condition.right_operand())

        count_for_each_val = [cp.Count(cpm_vars, val) for val in cpm_vals]
        self.cpm_model += self.eval_cpm_comp(cp.sum(count_for_each_val), condition.operator, rhs)

    def ctr_atleast(self, lst: list[Variable] | list[Node], value: int, k: int):
        cpm_exprs = self.get_cpm_exprs(lst)
        self.cpm_model += (cp.Count(cpm_exprs, value) >= k)

    def ctr_atmost(self, lst: list[Variable], value: int, k: int):
        cpm_vars = self.get_cpm_vars(lst)
        self.cpm_model += (cp.Count(cpm_vars, value) <= k)

    def ctr_exactly(self, lst: list[Variable], value: int, k: int | Variable):
        cpm_vars = self.get_cpm_exprs(lst)
        self.cpm_model += (cp.Count(cpm_vars, value) == self.get_cpm_var(k))

    def ctr_among(self, lst: list[Variable], values: list[int], k: int | Variable):
        self.cpm_model += cp.Among(self.get_cpm_vars(lst), values) == self.get_cpm_var(k)

    def ctr_nvalues(self, lst: list[Variable] | list[Node], excepting: None | list[int], condition: Condition):
        if excepting is None:
            lhs = cp.NValue(self.get_cpm_exprs(lst))
        else:
            assert len(excepting) == 1, "Competition only allows 1 integer value in excepting list"
            lhs = cp.NValueExcept(self.get_cpm_exprs(lst), excepting[0])
        
        if condition.operator == TypeConditionOperator.IN or (condition.operator == TypeConditionOperator.NOTIN):
            from pycsp3.classes.auxiliary.conditions import ConditionInterval, ConditionSet
            assert isinstance(condition,
                              (ConditionInterval, ConditionSet)), "Competition only supports intervals when operator is `in` or `notin`"  # TODO and not in? (notin)
            if isinstance(condition, ConditionInterval):
                rhs = list(range(condition.min, condition.max + 1))
            else: # ConditionSet
                rhs = list(condition.t)
        else:
            rhs = self.get_cpm_var(condition.right_operand())

        self.cpm_model += self.eval_cpm_comp(lhs, condition.operator, rhs)

    def ctr_not_all_qual(self, lst: list[Variable]):
        cpm_vars = self.get_cpm_vars(lst)
        self.cpm_model += cp.NValue(cpm_vars) > 1

    def ctr_cardinality(self, lst: list[Variable], values: list[int] | list[Variable],
                        occurs: list[int] | list[Variable] | list[range], closed: bool):
        self.cpm_model += cp.GlobalCardinalityCount(self.get_cpm_exprs(lst),
                                                    self.get_cpm_exprs(values),
                                                    self.get_cpm_exprs(occurs),
                                                    closed=closed)

    def ctr_minimum(self, lst: list[Variable] | list[Node], condition: Condition):
        cpm_vars = self.get_cpm_exprs(lst)
        self.cpm_model += self.eval_cpm_comp(cp.Minimum(cpm_vars),
                                             condition.operator,
                                             self.get_cpm_var(condition.right_operand()))

    def ctr_maximum(self, lst: list[Variable] | list[Node], condition: Condition):
        cpm_vars = self.get_cpm_exprs(lst)
        self.cpm_model += self.eval_cpm_comp(cp.Maximum(cpm_vars),
                                             condition.operator,
                                             self.get_cpm_var(condition.right_operand()))

    def ctr_minimum_arg(self, lst: list[Variable] | list[Node], condition: Condition,
                        rank: TypeRank):  # should enter XCSP3-core
        self._unimplemented(lst, condition, rank)

    def ctr_maximum_arg(self, lst: list[Variable] | list[Node], condition: Condition,
                        rank: TypeRank):  # should enter XCSP3-core
        self._unimplemented(lst, condition, rank)

    def ctr_element(self, lst: list[Variable] | list[int], i: Variable, condition: Condition):
        cpm_lst = self.get_cpm_vars(lst)
        cpm_index = self.get_cpm_var(i)
        cpm_rhs = self.get_cpm_var(condition.right_operand())
        self.cpm_model += self.eval_cpm_comp(cp.Element(cpm_lst, cpm_index), condition.operator, cpm_rhs)

    def ctr_element_matrix(self, matrix: list[list[Variable]] | list[list[int]], i: Variable, j: Variable,
                           condition: Condition):
        
        # this can be optimized by indexing into the matrix directly
        mtrx = cp.cpm_array([self.get_cpm_vars(lst) for lst in matrix])
        dim1, dim2 = mtrx.shape

        cpm_i, cpm_j = self.get_cpm_vars([i, j])
        cpm_rhs = self.get_cpm_var(condition.right_operand())
        # ensure i,j are within bounds, we can do this as it is a toplevel constraint
        self.cpm_model += [cpm_i >= 0, cpm_i < dim1, cpm_j >= 0, cpm_j < dim2]

        # flatten matrix and lookup with weighed sum
        self.cpm_model += self.eval_cpm_comp(cp.Element(flatlist(mtrx), dim1 * cpm_i + cpm_j), condition.operator, cpm_rhs)


    def ctr_channel(self, lst1: list[Variable], lst2: None | list[Variable]):

        if lst2 is None:
            self.cpm_model += xglobals.InverseOne(self.get_cpm_vars(lst1))
        else:
            cpm_vars1 = self.get_cpm_vars(lst1)
            cpm_vars2 = self.get_cpm_vars(lst2)
            # Ignace: deprecated, we have InverseAsym now which gets constructed automatically
            # # make lists same length, last part is irrelevant if not same length
            # if len(cpm_vars2) > len(cpm_vars1):
            #     cpm_vars2 = cpm_vars2[0:len(cpm_vars1)]
            # elif len(cpm_vars1) > len(cpm_vars2):
            #     cpm_vars1 = cpm_vars1[0:len(cpm_vars2)]
            self.cpm_model += xglobals.SafeOnlyInverse(cpm_vars1, cpm_vars2)

    def ctr_channel_value(self, lst: list[Variable], value: Variable):
        self.cpm_model += xglobals.Channel(self.get_cpm_vars(lst), self.get_cpm_var(value))

    def ctr_nooverlap(self, origins: list[Variable], lengths: list[int] | list[Variable],
                      zero_ignored: bool):  # in XCSP3 competitions, no 0 permitted in lengths
        cpm_start = self.get_cpm_vars(origins)
        cpm_dur = self.get_cpm_vars(lengths)
        cpm_end = [cp.intvar(*get_bounds(s + d)) for s, d in zip(cpm_start, cpm_dur)]
        self.cpm_model += cp.NoOverlap(cpm_start, cpm_dur, cpm_end)

    def ctr_nooverlap_multi(self, origins: list[list[Variable]], lengths: list[list[int]] | list[list[Variable]],
                            zero_ignored: bool):
        dim = len(origins[0])
        if dim == 2:

            start_x, start_y = self.get_cpm_vars([o[0] for o in origins]), self.get_cpm_vars([o[1] for o in origins])
            dur_x, dur_y = self.get_cpm_vars([l[0] for l in lengths]), self.get_cpm_vars([l[1] for l in lengths])

            end_x = [cp.intvar(*get_bounds(s + d)) for s, d in zip(start_x, dur_x)]
            end_y = [cp.intvar(*get_bounds(s + d)) for s, d in zip(start_y, dur_y)]

            self.cpm_model += xglobals.NoOverlap2d(start_x, dur_x, end_x,
                                             start_y, dur_y, end_y)

        else:  # n-dimensional, post decomposition directly
            from cpmpy.expressions.utils import all_pairs
            from cpmpy import any as cpm_any
            starts = cp.cpm_array([self.get_cpm_vars(lst) for lst in origins])
            durs = cp.cpm_array([self.get_cpm_vars(lst) for lst in lengths])

            for i, j in all_pairs(list(range(len(origins)))):
                self.cpm_model += cpm_any([(starts[i, d] + durs[i, d] <= starts[j, d]) | \
                                           (starts[j, d] + durs[j, d] <= starts[i, d]) for d in range(dim)])

    def ctr_nooverlap_mixed(self, xs: list[Variable], ys: list[Variable], lx: list[Variable], ly: list[int],
                            zero_ignored: bool):
        start_x = self.get_cpm_vars(xs)
        start_y = self.get_cpm_vars(ys)
        dur_x = self.get_cpm_vars(lx)
        dur_y = ly

        end_x = [cp.intvar(*get_bounds(s + d)) for s, d in zip(start_x, dur_x)]
        end_y = [cp.intvar(*get_bounds(s + d)) for s, d in zip(start_y, dur_y)]

        self.cpm_model += xglobals.NoOverlap2d(start_x, dur_x, end_x,
                                         start_y, dur_y, end_y)

    def ctr_cumulative(self, origins: list[Variable], lengths: list[int] | list[Variable],
                       heights: list[int] | list[Variable], condition: Condition):
        cpm_start = self.get_cpm_exprs(origins)
        cpm_durations = self.get_cpm_exprs(lengths)
        cpm_demands = self.get_cpm_exprs(heights)
        cpm_ends = []
        for s, d in zip(cpm_start, cpm_durations):
            expr = s + d
            cpm_ends.append(cp.intvar(*get_bounds(expr)))

        if condition.operator == TypeConditionOperator.LE:
            self.cpm_model += xglobals.DynamicCumulative(cpm_start, cpm_durations, cpm_ends, cpm_demands,
                                            self.get_cpm_var(condition.right_operand()))
        else:
            # post decomposition directly
            # be smart and chose task or time decomposition #TODO you did task decomp in both cases
            if max(get_bounds(cpm_ends)[1]) >= 100:
                self._cumulative_task_decomp(cpm_start, cpm_durations, cpm_ends, heights, condition)
            else:
                self._cumulative_time_decomp(cpm_start, cpm_durations, cpm_ends, heights, condition)

    def _cumulative_task_decomp(self, cpm_start, cpm_duration, cpm_ends, cpm_demands, condition: Condition):
        cpm_demands = cp.cpm_array(cpm_demands)
        cpm_cap = self.get_cpm_var(condition.right_operand())
        # ensure durations are satisfied
        for s, d, e in zip(cpm_start, cpm_duration, cpm_ends):
            self.cpm_model += s + d == e

        # task decomposition
        for s, d, e in zip(cpm_start, cpm_duration, cpm_ends):
            # find overlapping tasks
            total_running = cp.sum(cpm_demands * ((cpm_start <= s) & (cpm_ends > s)))
            self.cpm_model += self.eval_cpm_comp(total_running, condition.operator, cpm_cap)

    def _cumulative_time_decomp(self, cpm_start, cpm_duration, cpm_ends, cpm_demands, condition: Condition):
        cpm_demands = cp.cpm_array(cpm_demands)
        cpm_cap = self.get_cpm_var(condition.right_operand)
        # ensure durations are satisfied
        for s, d, e in zip(cpm_start, cpm_duration, cpm_ends):
            self.cpm_model += s + d == e

        lb = min(get_bounds(cpm_start)[0])
        ub = max(get_bounds(cpm_ends)[1])
        # time decomposition
        for t in range(lb, ub + 1):
            total_running = cp.sum(cpm_demands * ((cpm_start <= t) & (cpm_ends > t)))
            self.cpm_model += self.eval_cpm_comp(total_running, condition.operator, cpm_cap)

    def ctr_binpacking(self, lst: list[Variable], sizes: list[int], condition: Condition):
        cpm_vars = self.get_cpm_vars(lst)
        cpm_rhs = self.get_cpm_var(condition.right_operand())

        for bin in range(0, len(cpm_vars)):  # bin labeling starts at 0, contradicting the xcsp3 specification document?
            self.cpm_model += self.eval_cpm_comp(cp.sum((cpm_array(cpm_vars) == bin) * sizes),
                                                 condition.operator,
                                                 cpm_rhs)

    def ctr_binpacking_limits(self, lst: list[Variable], sizes: list[int], limits: list[int] | list[Variable]):
        from cpmpy.expressions.utils import eval_comparison

        cpm_vars = self.get_cpm_vars(lst)

        for bin, lim in enumerate(limits):
            self.cpm_model += eval_comparison("<=",
                                              cp.sum((cpm_array(cpm_vars) == (bin)) * sizes),
                                              lim)

    def ctr_binpacking_loads(self, lst: list[Variable], sizes: list[int], loads: list[int] | list[Variable]):
        from cpmpy.expressions.utils import eval_comparison

        cpm_vars = self.get_cpm_vars(lst)
        cpm_loads = self.get_cpm_vars(loads)

        for bin, load in enumerate(cpm_loads):
            self.cpm_model += eval_comparison("==",
                                              cp.sum((cpm_array(cpm_vars) == (bin)) * sizes),
                                              load)

    def ctr_binpacking_conditions(self, lst: list[Variable], sizes: list[int],
                                  conditions: list[Condition]):  # not in XCSP3-core
        self._unimplemented(lst, sizes, conditions)

    def ctr_knapsack(self, lst: list[Variable], weights: list[int], wcondition: Condition, profits: list[int],
                     pcondition: Condition):

        vars = cpm_array(self.get_cpm_vars(lst))
        cpm_weight = self.get_cpm_var(wcondition.right_operand())
        cpm_profit = self.get_cpm_var(pcondition.right_operand())

        total_weight = cp.sum(vars * weights)
        total_profit = cp.sum(vars * profits)
        self.cpm_model += self.eval_cpm_comp(total_weight, wcondition.operator, cpm_weight)
        self.cpm_model += self.eval_cpm_comp(total_profit, pcondition.operator, cpm_profit)

    def ctr_flow(self, lst: list[Variable], balance: list[int] | list[Variable], arcs: list,
                 capacities: None | list[range]):  # not in XCSP3-core
        self._unimplemented(lst, balance, arcs, capacities)

    def ctr_flow_weighted(self, lst: list[Variable], balance: list[int] | list[Variable], arcs: list,
                          capacities: None | list[range],
                          weights: list[int] | list[Variable],
                          condition: Condition):  # not in XCSP3-core
        self._unimplemented(lst, balance, arcs, capacities, weights, condition)

    def ctr_instantiation(self, lst: list[Variable], values: list[int]):
        self.cpm_model += xglobals.NonReifiedTable(self.get_cpm_vars(lst), [values])

    def ctr_clause(self, pos: list[Variable], neg: list[Variable]):  # not in XCSP3-core
        self._unimplemented(pos, neg)

    def ctr_circuit(self, lst: list[Variable], size: None | int | Variable):  # size is None in XCSP3 competitions
        self.cpm_model += xglobals.SubCircuitWithStart(self.get_cpm_vars(lst), start_index=0)

    # # # # # # # # # #
    # All methods about objectives to be implemented
    # # # # # # # # # #

    def obj_minimize(self, term: Variable | Node):
        if isinstance(term, Node):
            cpm_expr = self.exprs_from_node([term])
            assert len(cpm_expr) == 1
            cpm_expr = cpm_expr[0]
        else:
            cpm_expr = self.get_cpm_var(term)
        self.cpm_model.minimize(cpm_expr)

    def obj_maximize(self, term: Variable | Node):
        if isinstance(term, Node):
            cpm_expr = self.exprs_from_node([term])
            assert len(cpm_expr) == 1
            cpm_expr = cpm_expr[0]
        else:
            cpm_expr = self.get_cpm_var(term)
        self.cpm_model.maximize(cpm_expr)

    def obj_minimize_special(self, obj_type: TypeObj, terms: list[Variable] | list[Node],
                             coefficients: None | list[int]):
        import numpy as np
        if coefficients is None:
            coefficients = np.ones(len(terms))
        else:
            coefficients = np.array(coefficients)

        if obj_type == TypeObj.SUM:
            self.cpm_model.minimize(cp.sum(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.MAXIMUM:
            self.cpm_model.minimize(cp.max(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.MINIMUM:
            self.cpm_model.minimize(cp.min(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.NVALUES:
            self.cpm_model.minimize(cp.NValue(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.EXPRESSION:
            assert all(coeff == 1 for coeff in coefficients)
            assert len(terms) == 1
            cpm_expr = self.get_cpm_exprs(terms)[0]
            self.cpm_model.minimize(cpm_expr)
        else:
            self._unimplemented(obj_type, coefficients, terms)

    # TODO objectives are a bit confusing but i think it's correct
    def obj_maximize_special(self, obj_type: TypeObj, terms: list[Variable] | list[Node],
                             coefficients: None | list[int]):
        import numpy as np
        if coefficients is None:
            coefficients = np.ones(len(terms))
        else:
            coefficients = np.array(coefficients)

        if obj_type == TypeObj.SUM:
            self.cpm_model.maximize(cp.sum(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.MAXIMUM:
            self.cpm_model.maximize(cp.max(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.MINIMUM:
            self.cpm_model.maximize(cp.min(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.NVALUES:
            self.cpm_model.maximize(cp.NValue(coefficients * self.get_cpm_exprs(terms)))
        elif obj_type == TypeObj.EXPRESSION:
            assert all(coeff == 1 for coeff in coefficients)
            assert len(terms) == 1
            cpm_expr = self.get_cpm_exprs(terms)[0]
            self.cpm_model.maximize(cpm_expr)
        else:
            self._unimplemented(obj_type, coefficients, terms)

    def vars_from_node(self, scope):
        cpm_vars = []
        for var in scope:
            cpm_var = self.cpm_variables[var]
            cpm_vars.append(cpm_var)
        return cpm_vars

    def exprs_from_node(self, node):
        cpm_exprs = []
        for expr in node:
            cpm_expr = self.intentionfromtree(expr)
            cpm_exprs.append(cpm_expr)
        return cpm_exprs

    def get_cpm_var(self, x):
        if isinstance(x, XVar):
            return self.cpm_variables[x]
        else:
            return x  # constants

    def get_cpm_vars(self, lst):
        if isinstance(lst[0], (XVar, int)):
            return [self.get_cpm_var(x) for x in lst]
        if isinstance(lst[0], range):
            assert len(lst) == 1, "Expected range here, but got list with multiple elements, what's the semantics???"
            return list(lst[0])  # this should work without converting to str first
            # return list(eval(str(lst[0])))
        else:
            return self.vars_from_node(lst)

    def get_cpm_exprs(self, lst):
        if isinstance(lst[0], XVar):
            return [self.get_cpm_var(x) for x in lst]
        if isinstance(lst[0], range):
            # assert len(lst) == 1, f"Expected range here, but got list with multiple elements, what's the semantics???{lst}"

            if len(lst) == 1:
                return list(lst[0])  # this should work without converting to str first
            else:
                return [cp.intvar(l.start, l.stop - 1) for l in lst]

            # return list(eval(str(lst[0])))
        else:
            return self.exprs_from_node(lst)

    def end_instance(self):
        pass

    def load_annotation(self, annotation):
        pass

    def load_annotations(self, annotations):
        pass

    def load_objectives(self, objectives):
        pass

    def ann_decision(self, lst: list[Variable]):
        pass

    def ann_val_heuristic_static(self, lst: list[Variable], order: list[int]):
        pass



"""
PyTorch-style Dataset for XCSP3 competition instances.

Simply create a dataset instance (configured for the targeted competition year/track) and start iterating over its contents:

.. code-block:: python

    from cpmpy.tools.xcsp3 import XCSP3Dataset, read_xcsp3

    for filename, metadata in XCSP3Dataset(year=2024, track="COP", download=True): # auto download dataset and iterate over its instances
        # Do whatever you want here, e.g. reading to a CPMpy model and solving it:
        model = read_xcsp3(filename)
        model.solve()
        print(model.status())

The `metadata` contains usefull information about the current problem instance.

Since the dataset is PyTorch compatible, it can be used with a DataLoader:

.. code-block:: python

    from cpmpy.tools.xcsp3 import XCSP3Dataset, read_xcsp3

    # Initialize the dataset
    dataset = XCSP3Dataset(year=2024, track="COP", download=True)

    from torch.utils.data import DataLoader

    # Wrap dataset in a DataLoader
    data_loader = DataLoader(dataset, batch_size=10, shuffle=False)

    # Iterate over the dataset
    for batch in data_loader:
        # Your code here
"""

import pathlib
from typing import Tuple, Any
import xml.etree.ElementTree as ET
from urllib.request import urlretrieve
from urllib.error import HTTPError, URLError
import zipfile

class XCSP3Dataset(object):  # torch.utils.data.Dataset compatible

    """
    XCSP3 Dataset in a PyTorch compatible format.
    
    Arguments:
        root (str): Root directory containing the XCSP3 instances (if 'download', instances will be downloaded to this location)
        year (int): Competition year (2022, 2023 or 2024)
        track (str, optional): Filter instances by track type (e.g., "COP", "CSP", "MiniCOP")
        transform (callable, optional): Optional transform to be applied on the instance data (the file path of each problem instance)
        target_transform (callable, optional): Optional transform to be applied on the metadata (the metadata dictionary of each problem instance)
        download (bool): If True, downloads the dataset from the internet and puts it in `root` directory
    """
    
    def __init__(self, root: str = ".", year: int = 2023, track: str = None, transform=None, target_transform=None, download: bool = False):
        """
        Initialize the XCSP3 Dataset.
        """
        self.root = pathlib.Path(root)
        self.year = year
        self.transform = transform
        self.target_transform = target_transform
        self.track = track
        self.track_dir = self.root / str(year) / track
        
        if not str(year).startswith('20'):
            raise ValueError("Year must start with '20'")
        if not track:
            raise ValueError("Track must be specified, e.g. COP, CSP, MiniCOP, ...")
        # Create root directory if it doesn't exist
        self.root.mkdir(parents=True, exist_ok=True)
        
        if not self.track_dir.exists():
            if not download:
                raise ValueError(f"Dataset for year {year} and track {track} not found. Please set download=True to download the dataset.")
            else:
                print(f"Downloading XCSP3 {year} instances...")
                url = f"https://www.cril.univ-artois.fr/~lecoutre/compets/"
                year_suffix = str(year)[2:]  # Drop the starting '20'
                url_path = url + f"instancesXCSP{year_suffix}.zip"
                zip_path = self.root / f"instancesXCSP{year_suffix}.zip"
                
                try:
                    urlretrieve(url_path, str(zip_path))
                except (HTTPError, URLError) as e:
                    raise ValueError(f"No dataset available for year {year}. Error: {str(e)}")
                
                # Extract only the specific track folder from the zip
                with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                    # Get the main folder name (e.g., "024_V3")
                    main_folder = None
                    for name in zip_ref.namelist():
                        if '/' in name:
                            main_folder = name.split('/')[0]
                            break
                    
                    if main_folder is None:
                        raise ValueError(f"Could not find main folder in zip file")

                    # Extract only files from the specified track
                    # Get all unique track names from zip
                    tracks = set()
                    for file_info in zip_ref.infolist():
                        parts = file_info.filename.split('/')
                        if len(parts) > 2 and parts[0] == main_folder:
                            tracks.add(parts[1])
                    
                    # Check if requested track exists
                    if track not in tracks:
                        raise ValueError(f"Track '{track}' not found in dataset. Available tracks: {sorted(tracks)}")
                    
                    # Create track folder in root directory, parents=True ensures recursive creation
                    self.track_dir.mkdir(parents=True, exist_ok=True)
                    
                    # Extract files for the specified track
                    prefix = f"{main_folder}/{track}/"
                    for file_info in zip_ref.infolist():
                        if file_info.filename.startswith(prefix):
                            # Extract file to track_dir, removing main_folder/track prefix
                            filename = pathlib.Path(file_info.filename).name
                            with zip_ref.open(file_info) as source, open(self.track_dir / filename, 'wb') as target:
                                target.write(source.read())
                # Clean up the zip file
                zip_path.unlink()

        
    def __len__(self) -> int:
        """Return the total number of instances."""
        return len(list(self.track_dir.glob("*.xml.lzma")))
    
    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Get a single XCSP3 instance filename and metadata.

        Args:
            index (int): Index of the instance to retrieve
            
        Returns:
            Tuple[Any, Any]: A tuple containing:
                - The filename of the instance
                - Metadata dictionary with file name, track, year etc.
        """
        if index < 0 or index >= len(self):
            raise IndexError("Index out of range")

        # Get all compressed XML files and sort for deterministic behavior
        files = sorted(list(self.track_dir.glob("*.xml.lzma")))
        file_path = files[index]

        filename = str(file_path)
        if self.transform:
            # does not need to remain a filename...
            filename = self.transform(filename)
            
        # Basic metadata about the instance
        metadata = {
            'year': self.year,
            'track': self.track,
            'name': file_path.stem.replace('.xml.lzma', ''),
            'path': filename,
        }
        if self.target_transform:
            metadata = self.target_transform(metadata)
            
        return filename, metadata

if __name__ == "__main__":
    dataset = XCSP3Dataset(year=2024, track="MiniCOP", download=True)
    print("Dataset size:", len(dataset))
    print("Instance 0:", dataset[0])



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## gurobi.py
##
"""
    Interface to Gurobi Optimizer's Python API.

    Gurobi Optimizer is a highly efficient commercial solver for Integer Linear Programming (and more).

    Always use :func:`cp.SolverLookup.get("gurobi") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'gurobipy' python package is installed:

    .. code-block:: console

        $ pip install gurobipy
    
    Gurobi Optimizer requires an active licence (for example a free academic license)
    You can read more about available licences at https://www.gurobi.com/downloads/

    See detailed installation instructions at:
    https://support.gurobi.com/hc/en-us/articles/360044290292-How-do-I-install-Gurobi-for-Python-

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_gurobi

    ==============
    Module details
    ==============
"""

from typing import Optional
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import *
from ..expressions.utils import argvals, argval
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl, intvar
from ..expressions.globalconstraints import DirectConstraint
from ..transformations.comparison import only_numexpr_equality
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.flatten_model import flatten_constraint, flatten_objective
from ..transformations.get_variables import get_variables
from ..transformations.linearize import linearize_constraint, only_positive_bv, only_positive_bv_wsum
from ..transformations.normalize import toplevel_list
from ..transformations.reification import only_implies, reify_rewrite, only_bv_reifies
from ..transformations.safening import no_partial_functions

try:
    import gurobipy as gp
    GRB_ENV = None
except ImportError:
    pass


class CPM_gurobi(SolverInterface):
    """
    Interface to Gurobi's Python API

    Creates the following attributes (see parent constructor for more):
    
    - ``grb_model``: object, TEMPLATE's model object

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, calls a function on the ``grb_model`` object.
    
    Documentation of the solver's own Python API:
    https://docs.gurobi.com/projects/optimizer/en/current/reference/python.html
    """

    @staticmethod
    def supported():
        return CPM_gurobi.installed() and CPM_gurobi.license_ok()

    @staticmethod
    def installed():
        try:
            import gurobipy as gp
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e

    @staticmethod
    def license_ok():
        if not CPM_gurobi.installed():
            warnings.warn(f"License check failed, python package 'gurobipy' is not installed! Please check 'CPM_gurobi.installed()' before attempting to check license.")
            return False
        try:
            import gurobipy as gp
            global GRB_ENV
            if GRB_ENV is None:
                # initialise the native gurobi model object
                GRB_ENV = gp.Env(params={"OutputFlag": 0})
                GRB_ENV.start()
            return True
        except Exception as e:
            warnings.warn(f"Problem encountered with Gurobi license: {e}")
            return False
        
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('gurobipy').version
        except pkg_resources.DistributionNotFound:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
            cpm_model: a CPMpy Model()
            subsolver: None, not used
        """
        if not self.installed():
            raise Exception("CPM_gurobi: Install the python package 'gurobipy' to use this solver interface.")
        elif not self.license_ok():
            raise Exception("CPM_gurobi: A problem occured during license check. Make sure your license is activated!")
        import gurobipy as gp

        # TODO: subsolver could be a GRB_ENV if a user would want to hand one over
        self.grb_model = gp.Model(env=GRB_ENV)

        # initialise everything else and post the constraints/objective
        # it is sufficient to implement add() and minimize/maximize() below
        super().__init__(name="gurobi", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.grb_model


    def solve(self, time_limit=None, solution_callback=None, **kwargs):
        """
            Call the gurobi solver

            Arguments:
                time_limit (float, optional):  maximum solve time in seconds 
                **kwargs:                      any keyword argument, sets parameters of solver object

            Arguments that correspond to solver parameters:
            Examples of gurobi supported arguments include:

            - ``Threads`` : int
            - ``MIPFocus`` : int
            - ``ImproveStartTime`` : bool
            - ``FlowCoverCuts`` : int

            For a full list of gurobi parameters, please visit https://www.gurobi.com/documentation/9.5/refman/parameters.html#sec:Parameters
        """
        from gurobipy import GRB

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)
        
        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            self.grb_model.setParam("TimeLimit", time_limit)

        # call the solver, with parameters
        for param, val in kwargs.items():
            self.grb_model.setParam(param, val)

        _ = self.grb_model.optimize(callback=solution_callback)
        grb_objective = self.grb_model.getObjective()

        grb_status = self.grb_model.Status

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.grb_model.runtime

        # translate exit status
        if grb_status == GRB.OPTIMAL:
            # COP
            if self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            # CSP
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif grb_status == GRB.INFEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif grb_status == GRB.TIME_LIMIT:
            if self.grb_model.SolCount == 0:
                # can be sat or unsat
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        else:  # another?
            raise NotImplementedError(
                f"Translation of gurobi status {grb_status} to CPMpy status not implemented")  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                solver_val = self.solver_var(cpm_var).X
                if cpm_var.is_bool():
                    cpm_var._value = solver_val >= 0.5
                else:
                    cpm_var._value = int(solver_val)
            # set _objective_value
            if self.has_objective():
                grb_obj_val = grb_objective.getValue()
                if round(grb_obj_val) == grb_obj_val: # it is an integer?:
                    self.objective_value_ = int(grb_obj_val)
                else: #  can happen with DirectVar or when using floats as coefficients
                    self.objective_value_ =  float(grb_obj_val)

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var): # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view. Should be eliminated in linearize
        if isinstance(cpm_var, NegBoolView):
            raise NotSupportedError("Negative literals should not be left as part of any equation. Please report.")

        # create if it does not exit
        if cpm_var not in self._varmap:
            from gurobipy import GRB
            if isinstance(cpm_var, _BoolVarImpl):
                revar = self.grb_model.addVar(vtype=GRB.BINARY, name=cpm_var.name)
            elif isinstance(cpm_var, _IntVarImpl):
                revar = self.grb_model.addVar(cpm_var.lb, cpm_var.ub, vtype=GRB.INTEGER, name=str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        # return from cache
        return self._varmap[cpm_var]


    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            'objective()' can be called multiple times, only the last one is stored

            .. note::
                technical side note: any constraints created during conversion of the objective
                are premanently posted to the solver
        """
        from gurobipy import GRB

        # make objective function non-nested
        (flat_obj, flat_cons) = flatten_objective(expr, csemap=self._csemap)
        flat_obj = only_positive_bv_wsum(flat_obj)  # remove negboolviews
        get_variables(flat_obj, collect=self.user_vars)  # add potentially created variables
        self += flat_cons

        # make objective function or variable and post
        obj = self._make_numexpr(flat_obj)
        if minimize:
            self.grb_model.setObjective(obj, sense=GRB.MINIMIZE)
        else:
            self.grb_model.setObjective(obj, sense=GRB.MAXIMIZE)
        self.grb_model.update()

    def has_objective(self):
        return self.grb_model.getObjective().size() != 0  # TODO: check if better way to do this...

    def _make_numexpr(self, cpm_expr):
        """
            Turns a numeric CPMpy 'flat' expression into a solver-specific
            numeric expression

            Used especially to post an expression as objective function
        """
        import gurobipy as gp

        if is_num(cpm_expr):
            return cpm_expr

        # decision variables, check in varmap
        if isinstance(cpm_expr, _NumVarImpl):  # _BoolVarImpl is subclass of _NumVarImpl
            return self.solver_var(cpm_expr)

        # sum
        if cpm_expr.name == "sum":
            return gp.quicksum(self.solver_vars(cpm_expr.args))
        if cpm_expr.name == "sub":
            a,b = self.solver_vars(cpm_expr.args)
            return a - b
        # wsum
        if cpm_expr.name == "wsum":
            return gp.quicksum(w * self.solver_var(var) for w, var in zip(*cpm_expr.args))

        raise NotImplementedError("gurobi: Not a known supported numexpr {}".format(cpm_expr))


    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        # apply transformations, then post internally
        # expressions have to be linearized to fit in MIP model. See /transformations/linearize
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel={"mod", "div"})  # linearize expects safe exprs
        supported = {"min", "max", "abs", "alldifferent"} # alldiff has a specialized MIP decomp in linearize
        cpm_cons = decompose_in_tree(cpm_cons, supported, csemap=self._csemap)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['sum', 'wsum']), csemap=self._csemap)  # constraints that support reification
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum", "sub"]), csemap=self._csemap)  # supports >, <, !=
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)  # anything that can create full reif should go above...
        # gurobi does not round towards zero, so no 'div' in supported set: https://github.com/CPMpy/cpmpy/pull/593#issuecomment-2786707188
        cpm_cons = linearize_constraint(cpm_cons, supported=frozenset({"sum", "wsum","sub","min","max","mul","abs","pow"}), csemap=self._csemap)  # the core of the MIP-linearization
        cpm_cons = only_positive_bv(cpm_cons, csemap=self._csemap)  # after linearization, rewrite ~bv into 1-bv
        return cpm_cons

    def add(self, cpm_expr_orig):
      """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
      """
      from gurobipy import GRB

      # add new user vars to the set
      get_variables(cpm_expr_orig, collect=self.user_vars)

      # transform and post the constraints
      for cpm_expr in self.transform(cpm_expr_orig):

        # Comparisons: only numeric ones as 'only_implies()' has removed the '==' reification for Boolean expressions
        # numexpr `comp` bvar|const
        if isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            grbrhs = self.solver_var(rhs)

            # Thanks to `only_numexpr_equality()` only supported comparisons should remain
            if cpm_expr.name == '<=':
                grblhs = self._make_numexpr(lhs)
                self.grb_model.addLConstr(grblhs, GRB.LESS_EQUAL, grbrhs)
            elif cpm_expr.name == '>=':
                grblhs = self._make_numexpr(lhs)
                self.grb_model.addLConstr(grblhs, GRB.GREATER_EQUAL, grbrhs)
            elif cpm_expr.name == '==':
                if isinstance(lhs, _NumVarImpl) \
                        or (isinstance(lhs, Operator) and (lhs.name == 'sum' or lhs.name == 'wsum' or lhs.name == "sub")):
                    # a BoundedLinearExpression LHS, special case, like in objective
                    grblhs = self._make_numexpr(lhs)
                    self.grb_model.addLConstr(grblhs, GRB.EQUAL, grbrhs)

                elif lhs.name == 'mul':
                    assert len(lhs.args) == 2, "Gurobi only supports multiplication with 2 variables"
                    a, b = self.solver_vars(lhs.args)
                    self.grb_model.setParam("NonConvex", 2)
                    self.grb_model.addConstr(a * b == grbrhs)

                elif lhs.name == 'div':
                    if not is_num(lhs.args[1]):
                        raise NotSupportedError(f"Gurobi only supports division by constants, but got {lhs.args[1]}")
                    a, b = self.solver_vars(lhs.args)
                    self.grb_model.addLConstr(a / b, GRB.EQUAL, grbrhs)

                else:
                    # General constraints
                    # grbrhs should be a variable for gurobi in the subsequent, fake it
                    if is_num(grbrhs):
                        grbrhs = self.solver_var(intvar(lb=grbrhs, ub=grbrhs))

                    if lhs.name == 'min':
                        self.grb_model.addGenConstrMin(grbrhs, self.solver_vars(lhs.args))
                    elif lhs.name == 'max':
                        self.grb_model.addGenConstrMax(grbrhs, self.solver_vars(lhs.args))
                    elif lhs.name == 'abs':
                        self.grb_model.addGenConstrAbs(grbrhs, self.solver_var(lhs.args[0]))
                    elif lhs.name == 'pow':
                        x, a = self.solver_vars(lhs.args)
                        self.grb_model.addGenConstrPow(x, grbrhs, a)
                    else:
                        raise NotImplementedError(
                        "Not a known supported gurobi comparison '{}' {}".format(lhs.name, cpm_expr))
            else:
                raise NotImplementedError(
                "Not a known supported gurobi comparison '{}' {}".format(lhs.name, cpm_expr))

        elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            # Indicator constraints
            # Take form bvar -> sum(x,y,z) >= rvar
            cond, sub_expr = cpm_expr.args
            assert isinstance(cond, _BoolVarImpl), f"Implication constraint {cpm_expr} must have BoolVar as lhs"
            assert isinstance(sub_expr, Comparison), "Implication must have linear constraints on right hand side"
            if isinstance(cond, NegBoolView):
                cond, bool_val = self.solver_var(cond._bv), False
            else:
                cond, bool_val = self.solver_var(cond), True

            lhs, rhs = sub_expr.args
            if isinstance(lhs, _NumVarImpl) or lhs.name == "sum" or lhs.name == "wsum":
                lin_expr = self._make_numexpr(lhs)
            else:
                raise Exception(f"Unknown linear expression {lhs} on right side of indicator constraint: {cpm_expr}")
            if sub_expr.name == "<=":
                self.grb_model.addGenConstrIndicator(cond, bool_val, lin_expr, GRB.LESS_EQUAL, self.solver_var(rhs))
            elif sub_expr.name == ">=":
                self.grb_model.addGenConstrIndicator(cond, bool_val, lin_expr, GRB.GREATER_EQUAL, self.solver_var(rhs))
            elif sub_expr.name == "==":
                self.grb_model.addGenConstrIndicator(cond, bool_val, lin_expr, GRB.EQUAL, self.solver_var(rhs))
            else:
                raise Exception(f"Unknown linear expression {sub_expr} name")

        # True or False
        elif isinstance(cpm_expr, BoolVal):
            self.grb_model.addConstr(cpm_expr.args[0])

        # a direct constraint, pass to solver
        elif isinstance(cpm_expr, DirectConstraint):
            cpm_expr.callSolver(self, self.grb_model)

        else:
            raise NotImplementedError(cpm_expr)  # if you reach this... please report on github

      return self
    __add__ = add  # avoid redirect in superclass

    def solution_hint(self, cpm_vars, vals):
        """
        Gurobi supports warmstarting the solver with a (in)feasible solution.
        The provided value will affect branching heurstics during solving, making it more likely the final solution will contain the provided assignment.

        To learn more about solution hinting in gurobi, see:
        https://docs.gurobi.com/projects/optimizer/en/current/reference/attributes/variable.html#varhintval

        Optionally, you can also set the relative priority of the hint, using:
        
        .. code-block:: python

            solver.solver_var(cpm_var).setAttr("VarHintPri", <priority>)

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        for cpm_var, val in zip(cpm_vars, vals):
            self.solver_var(cpm_var).setAttr("VarHintVal", val)

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            Compute all solutions and optionally display the solutions.

            This is the generic implementation, solvers can overwrite this with
            a more efficient native implementation

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                time_limit: stop after this many seconds (default: None)
                solution_limit: stop after this many solutions (default: None)
                call_from_model: whether the method is called from a CPMpy Model instance or not
                any other keyword argument

            Returns: number of solutions found
        """
        from gurobipy import GRB

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)

        if time_limit is not None:
            self.grb_model.setParam("TimeLimit", time_limit)

        if solution_limit is None:
            raise Exception(
                "Gurobi does not support searching for all solutions. If you really need all solutions, "
                "try setting solution limit to a large number")

        # Force gurobi to keep searching in the tree for optimal solutions
        sa_kwargs = {"PoolSearchMode":2, "PoolSolutions":solution_limit}

        # solve the model
        self.solve(time_limit=time_limit, **sa_kwargs, **kwargs)

        optimal_val = None
        solution_count = self.grb_model.SolCount
        opt_sol_count = 0

        # clear user vars if no solution found
        if solution_count == 0:
            self.objective_value_ = None
            for var in self.user_vars:
                var._value = None

        for i in range(solution_count):
            # Specify which solution to query
            self.grb_model.setParam("SolutionNumber", i)
            sol_obj_val = self.grb_model.PoolObjVal
            if optimal_val is None:
                optimal_val = sol_obj_val
            if optimal_val is not None:
                # sub-optimal solutions
                if sol_obj_val != optimal_val:
                    break
            opt_sol_count += 1

            # Translate solution to variables
            for cpm_var in self.user_vars:
                solver_val = self.solver_var(cpm_var).Xn
                if cpm_var.is_bool():
                    cpm_var._value = solver_val >= 0.5
                else:
                    cpm_var._value = int(solver_val)

            # Translate objective
            if self.has_objective():
                self.objective_value_ = self.grb_model.PoolObjVal

            if display is not None:
                if isinstance(display, Expression):
                    print(argval(display))
                elif isinstance(display, list):
                    print(argvals(display))
                else:
                    display()  # callback

        # Reset pool search mode to default
        self.grb_model.setParam("PoolSearchMode", 0)

        if opt_sol_count:
            if opt_sol_count == solution_limit:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE 
            else:
                grb_status = self.grb_model.Status
                if grb_status == GRB.TIME_LIMIT: # reached time limit
                    self.cpm_status.exitstatus = ExitStatus.FEASIBLE
                else: # found all solutions   
                    self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        # if unsat or timout with no solution, .solve() will have already set the state accordingly (so nothing to update)

        return opt_sol_count



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## cplex.py
##
"""
    Interface to CPLEX Optimizer using the python 'docplex.mp' package

    CPLEX, standing as an acronym for ‘Complex Linear Programming Expert’,
    is a high-performance mathematical programming solver specializing in linear programming (LP),
    mixed integer programming (MIP), and quadratic programming (QP).

    Always use :func:`cp.SolverLookup.get("cplex") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============
    
    Requires that both the 'docplex' and the 'cplex' python packages are installed:

    .. code-block:: console

        $ pip install docplex cplex

    Detailed installation instructions available at:
    https://ibmdecisionoptimization.github.io/docplex-doc/getting_started_python.html

    You will also need to install CPLEX Optimization Studio from IBM's website.
    There is a free community version available.
    https://www.ibm.com/products/ilog-cplex-optimization-studio
    See detailed installation instructions at:
    https://www.ibm.com/docs/en/icos/22.1.2?topic=2212-installing-cplex-optimization-studio
    
    It also requires an active licence.
    Academic license:
    https://community.ibm.com/community/user/ai-datascience/blogs/xavier-nodet1/2020/07/09/cplex-free-for-students

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_cplex

    ==============
    Module details
    ==============
"""
import warnings
from typing import Optional

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import *
from ..expressions.utils import argvals, argval, eval_comparison, flatlist, is_bool
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl, intvar
from ..expressions.globalconstraints import DirectConstraint
from ..transformations.comparison import only_numexpr_equality
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.flatten_model import flatten_constraint, flatten_objective
from ..transformations.get_variables import get_variables
from ..transformations.linearize import linearize_constraint, only_positive_bv, only_positive_bv_wsum
from ..transformations.normalize import toplevel_list
from ..transformations.reification import only_implies, reify_rewrite, only_bv_reifies
from ..transformations.safening import no_partial_functions

class CPM_cplex(SolverInterface):
    """
    Interface to the CPLEX solver.

    Creates the following attributes (see parent constructor for more):

    - cplex_model: object, CPLEX model object

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, 
    calls a function on the ``cplex_model`` object.

    Documentation of the solver's own Python API:
    https://ibmdecisionoptimization.github.io/docplex-doc/mp/docplex.mp.model.html
    """

    @staticmethod
    def supported():
        return CPM_cplex.installed() and CPM_cplex.license_ok()

    @staticmethod
    def installed():
        # try to import the package
        try:
            import docplex.mp as domp
        except ModuleNotFoundError as e:
            warnings.warn(f"CPM_cplex: Could not import docplex: {e}")
            return False
        try:
            import cplex
            return True
        except ModuleNotFoundError as e:
            warnings.warn(f"CPM_cplex: Could not import cplex: {e}")
            return False

    @staticmethod
    def license_ok():
        if not CPM_cplex.installed():
            warnings.warn(
                f"License check failed, python package 'docplex' or 'cplex' is not installed! Please check 'CPM_cplex.installed()' before attempting to check license.")
            return False
        else:
            try:
                from docplex.mp.model import Model
                mdl = Model()
                mdl.solve()
                return True
            except Exception as e:
                warnings.warn(f"Problem encountered with CPLEX installation: {e}")
                return False

    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        
        Two version numbers get returned: ``<docplex version>/<solver version>``
        """
        try:
            import pkg_resources
            import cplex
            cpx = cplex.Cplex()
            return f"{pkg_resources.get_distribution('docplex').version}/{cpx.get_version()}"
        except (pkg_resources.DistributionNotFound, ModuleNotFoundError):
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
        - cpm_model: a CPMpy Model()
        - subsolver: None, not used
        """
        if not self.installed():
            raise Exception("CPM_cplex: Install the python packages 'docplex' and 'cplex' to use this solver interface.")
        elif not self.license_ok():
            raise Exception("CPM_cplex: A problem occured during license check. Make sure your installed the CPLEX Optimization Studio and that you have an active license.")

        from docplex.mp.model import Model
        self.cplex_model = Model()
        super().__init__(name="cplex", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.cplex_model

    def solve(self, time_limit=None, **kwargs):
        """
            Call the cplex solver

            Arguments:
            - time_limit:  maximum solve time in seconds (float, optional)
            - kwargs:      any keyword argument, sets parameters of solver object and cplex parameters

            Supported keyword arguments are all solve parameters and cplex parameters:
                - solve_parameters:
                    - context (optional) – context to use during solve
                    - checker (optional) – a string which controls which type of checking is performed. (type checks etc.)
                    - log_output (optional) – if True, solver logs are output to stdout.
                    - clean_before_solve (optional) – default False (iterative solving)
                - cplex_parameters:
                    - any cplex parameter, see https://www.ibm.com/docs/en/icos/22.1.2?topic=cplex-list-parameters
                    - a well-know parameter is the `threads` parameter, used to set the number of threads to use during solve

            For a full description of the parameters, please visit https://ibmdecisionoptimization.github.io/docplex-doc/mp/docplex.mp.model.html?#docplex.mp.model.Model.solve

            After solving, all solve details can be accessed through self.cplex_model.solve_details:
            https://ibmdecisionoptimization.github.io/docplex-doc/mp/docplex.mp.sdetails.html#docplex.mp.sdetails.SolveDetails
        """
        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)
            
        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            self.cplex_model.set_time_limit(time_limit)
    
        # Handle special arguments
        solve_args = ["clean_before_solve", "checker", "log_output"]
        cplex_params = {}
        
        for arg in list(kwargs.keys()):
            if arg == "context":
                self.cplex_model.context = kwargs[arg]
                del kwargs[arg]
            elif arg not in solve_args:
                # Set as cplex parameter
                cplex_params[arg] = kwargs[arg] 
                del kwargs[arg]
        
        self.cplex_model.solve(cplex_parameters=cplex_params, **kwargs)
        
        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.cplex_model.solve_details.time

        # translate solver exit status to CPMpy exit status
        cplex_status = self.cplex_model.solve_details.status
        if cplex_status == "Feasible":
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif "infeasible" in cplex_status:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif cplex_status == "Unknown":
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        elif "optimal" in cplex_status:            
            if self.has_objective(): # COP
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else: # CSP
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif cplex_status == "JobFailed":
            self.cpm_status.exitstatus = ExitStatus.ERROR
        elif "aborted" in cplex_status:
            self.cpm_status.exitstatus = ExitStatus.NOT_RUN
        else:  # another? This can happen when error during solve.
            raise NotImplementedError(f"Translation of cplex status {cplex_status} to CPMpy status not implemented")

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                solver_val = self.solver_var(cpm_var).solution_value
                if cpm_var.is_bool():
                    cpm_var._value = solver_val >= 0.5
                else:
                    cpm_var._value = round(solver_val)
            # set _objective_value
            if self.has_objective():
                obj_val = self.cplex_model.get_objective_expr().solution_value
                if round(obj_val) == obj_val: # it is an integer?:
                    self.objective_value_ = int(obj_val)
                else: #  can happen with DirectVar or when using floats as coefficients
                    self.objective_value_ = float(obj_val)

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var):  # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view
        if isinstance(cpm_var, NegBoolView):
            raise ValueError("Negative literals should not be part of any equation. "
                            "Should have been removed by the only_positive_bv() transformation. "
                            "See /transformations/linearize for more details")

        # create if it does not exit
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                revar = self.cplex_model.binary_var(cpm_var.name)
            elif isinstance(cpm_var, _IntVarImpl):
                revar = self.cplex_model.integer_var(cpm_var.lb, cpm_var.ub, name=str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        # return from cache
        return self._varmap[cpm_var]


    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            'objective()' can be called multiple times, only the last one is stored
            
            .. note::
                technical side note: any constraints created during conversion of the objective
                are premanently posted to the solver
        """
        # make objective function non-nested
        (flat_obj, flat_cons) = flatten_objective(expr)
        flat_obj = only_positive_bv_wsum(flat_obj)  # remove negboolviews
        get_variables(flat_obj, collect=self.user_vars)  # add potentially created variables
        self.add(flat_cons)

        # make objective function or variable and post
        obj = self._make_numexpr(flat_obj)
        if minimize:
            self.cplex_model.set_objective('min',obj)
        else:
            self.cplex_model.set_objective('max', obj)

    def has_objective(self):
        return self.cplex_model.is_optimized()

    def _make_numexpr(self, cpm_expr):
        """
            Turns a numeric CPMpy 'flat' expression into a solver-specific
            numeric expression

            Used especially to post an expression as objective function
        """
        if is_num(cpm_expr):
            return cpm_expr

        # decision variables, check in varmap
        if isinstance(cpm_expr, _NumVarImpl):  # _BoolVarImpl is subclass of _NumVarImpl
            return self.solver_var(cpm_expr)

        # sum
        if cpm_expr.name == "sum":
            return self.cplex_model.sum_vars(self.solver_vars(cpm_expr.args))

        # wsum
        if cpm_expr.name == "wsum":
            w, t = cpm_expr.args
            return self.cplex_model.scal_prod(self.solver_vars(t), w)

        if cpm_expr.name == "sub":
            a,b = self.solver_vars(cpm_expr.args)
            return a - b
        raise NotImplementedError("CPLEX: Not a known supported numexpr {}".format(cpm_expr))


    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: list of Expression
        """
        # apply transformations, then post internally
        # expressions have to be linearized to fit in MIP model. See /transformations/linearize
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel={"mod", "div"})  # linearize expects safe exprs
        supported = {"min", "max", "abs", "alldifferent"} # alldiff has a specialized MIP decomp in linearize
        cpm_cons = decompose_in_tree(cpm_cons, supported, csemap=self._csemap)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['sum', 'wsum', 'sub']), csemap=self._csemap)  # constraints that support reification
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum", "sub"]), csemap=self._csemap)  # supports >, <, !=
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)  # anything that can create full reif should go above...
        cpm_cons = linearize_constraint(cpm_cons, supported=frozenset({"sum", "wsum", "sub", "min", "max", "abs", "mul"}), csemap=self._csemap)  # CPLEX supports quadratic constraints and division by constants
        cpm_cons = only_positive_bv(cpm_cons, csemap=self._csemap)  # after linearization, rewrite ~bv into 1-bv
        return cpm_cons

    def add(self, cpm_expr_orig):
      """
        Eagerly add a constraint to the underlying solver.

        Any CPMpy expression given is immediately transformed (through `transform()`)
        and then posted to the solver in this function.

        This can raise `NotImplementedError` for any constraint not supported after transformation

        The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
        the user knows and cares about (and will be populated with a value after solve). All other variables
        are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
      """
      # add new user vars to the set
      get_variables(cpm_expr_orig, collect=self.user_vars)

      # transform and post the constraints
      for cpm_expr in self.transform(cpm_expr_orig):

        # Comparisons: only numeric ones as 'only_implies()' has removed the '==' reification for Boolean expressions
        # numexpr `comp` bvar|const
        if isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            cplexrhs = self.solver_var(rhs)

            # Thanks to `only_numexpr_equality()` only supported comparisons should remain
            if cpm_expr.name == '<=':
                cplexlhs = self._make_numexpr(lhs)
                self.cplex_model.add_constraint(cplexlhs <= cplexrhs)
            elif cpm_expr.name == '>=':
                cplexlhs = self._make_numexpr(lhs)
                self.cplex_model.add_constraint(cplexlhs >= cplexrhs)
            elif cpm_expr.name == '==':
                if isinstance(lhs, _NumVarImpl) \
                        or (isinstance(lhs, Operator) and (lhs.name in {'sum', 'wsum', 'sub'})):
                    # a BoundedLinearExpression LHS, special case, like in objective
                    cplexlhs = self._make_numexpr(lhs)
                    self.cplex_model.add_constraint(cplexlhs == cplexrhs)

                elif lhs.name == 'mul':
                    raise NotImplementedError(f'CPLEX only supports quadratic constraints that define a convex region, i.e. quadratic equalities are not supported: {cpm_expr}')

                else:
                    # Global functions
                    if lhs.name == 'min':
                        self.cplex_model.add_constraint(self.cplex_model.min(self.solver_vars(lhs.args)) == cplexrhs)
                    elif lhs.name == 'max':
                        self.cplex_model.add_constraint(self.cplex_model.max(self.solver_vars(lhs.args)) == cplexrhs)
                    elif lhs.name == 'abs':
                        self.cplex_model.add_constraint(self.cplex_model.abs(self.solver_var(lhs.args[0])) == cplexrhs)
                    else:
                        raise NotImplementedError(
                        "Not a known supported cplex comparison '{}' {}".format(lhs.name, cpm_expr))
            else:
                raise NotImplementedError(
                "Not a known supported cplex comparison '{}' {}".format(lhs.name, cpm_expr))

        elif isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            # Indicator constraints
            # Take form bvar -> sum(x,y,z) >= rvar
            cond, sub_expr = cpm_expr.args
            assert isinstance(cond, _BoolVarImpl), f"Implication constraint {cpm_expr} must have BoolVar as lhs"
            assert isinstance(sub_expr, Comparison), "Implication must have linear constraints on right hand side"
            if isinstance(cond, NegBoolView):
                cond, trigger_val = self.solver_var(cond._bv), False
            else:
                cond, trigger_val = self.solver_var(cond), True

            lhs, rhs = sub_expr.args
            if isinstance(lhs, _NumVarImpl) or (lhs.name in {'sum', 'wsum', 'sub'}):
                lin_expr = self._make_numexpr(lhs)
            else:
                raise ValueError(f"Unknown linear expression {lhs} on right side of indicator constraint: {cpm_expr}")
            constraint = eval_comparison(sub_expr.name, lin_expr, self.solver_var(rhs))
            self.cplex_model.add_indicator(cond, constraint, trigger_val)

        # True or False
        elif isinstance(cpm_expr, BoolVal):
            if cpm_expr.args[0]: # just true
                pass # do nothing
            else: # just false
                a = self.cplex_model.binary_var()
                self.cplex_model.add_constraint(a - a >= 1) # create a constraint that is always false

        # a direct constraint, pass to solver
        elif isinstance(cpm_expr, DirectConstraint):
            cpm_expr.callSolver(self, self.cplex_model)

        else:
            raise NotImplementedError(cpm_expr)  # if you reach this... please report on github

      return self
    __add__ = add  # avoid redirect in superclass

    def solution_hint(self, cpm_vars, vals):
        """
        CPLEX supports warmstarting the solver with a (in)feasible solution.
        This is done using MIP starts which provide the solver with a starting point
        for the branch-and-bound algorithm.

        The solution hint does NOT need to satisfy all constraints, it should just provide 
        reasonable default values for the variables. It can decrease solving times substantially, 
        especially when solving a similar model repeatedly.

        To learn more about solution hinting in CPLEX, see:
        https://ibmdecisionoptimization.github.io/docplex-doc/mp/docplex.mp.model.html#docplex.mp.model.Model.add_mip_start

        :param cpm_vars: list of CPMpy variables
        :param vals: list of (corresponding) values for the variables
        """
        # Flatten nested lists to handle test cases like solution_hint([a,[b]], [[[False]], True])
        cpm_vars = flatlist(cpm_vars)
        vals = flatlist(vals)
        
        # Validate input lengths
        if len(cpm_vars) != len(vals):
            raise ValueError(f"Number of variables ({len(cpm_vars)}) and values ({len(vals)}) must match")
        
        self.cplex_model.clear_mip_starts()

        # Create a MIP start solution using the proper docplex API
        if len(cpm_vars) > 0:
            warmstart = self.cplex_model.new_solution()
            for cpm_var, val in zip(cpm_vars, vals):
                # Convert boolean values to numeric (True -> 1, False -> 0) for docplex
                if is_bool(val):
                    val = int(val)
                warmstart.add_var_value(self.solver_var(cpm_var), val)

            self.cplex_model.add_mip_start(warmstart)

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            Compute all solutions and optionally display the solutions.

            This is the generic implementation, solvers can overwrite this with
            a more efficient native implementation

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                time_limit: stop after this many seconds (default: None)
                solution_limit: stop after this many solutions (default: None)
                any other keyword argument

            Returns: number of solutions found
        """
        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")
            self.cplex_model.set_time_limit(time_limit)

        if solution_limit is None:
            raise Exception(
                "CPLEX does not support searching for all solutions. If you really need all solutions, "
                "try setting solution limit to a large number")

        # Ask for multiple solutions
        self.cplex_model.context.cplex_parameters.mip.limits.populate = solution_limit
        self.cplex_model.context.cplex_parameters.mip.pool.intensity = 4 # (optional) max effort for finding solutions
        
        # For optimization problems, ensure we only get optimal solutions
        if self.has_objective():
            self.cplex_model.context.cplex_parameters.mip.pool.absgap = 0.0  # Only optimal solutions
            self.cplex_model.context.cplex_parameters.mip.pool.relgap = 0.0  # Only optimal solutions

        # Handle special arguments (same as in solve())
        solve_args = ["clean_before_solve", "checker", "log_output"]
        cplex_params = {}
        
        for arg in list(kwargs.keys()):
            if arg == "context":
                self.cplex_model.context = kwargs[arg]
                del kwargs[arg]
            elif arg not in solve_args:
                # Set as cplex parameter
                cplex_params[arg] = kwargs[arg] 
                del kwargs[arg]

        solutions_pool = self.cplex_model.populate_solution_pool(cplex_parameters=cplex_params, **kwargs)

        optimal_val = None
        opt_sol_count = 0

        # clear user vars if no solution found
        if solutions_pool is None:
            self.objective_value_ = None
            for var in self.user_vars:
                var._value = None

        else:
            for i, solution in enumerate(solutions_pool):
                if i == solution_limit:
                    break

                sol_obj_val = solution.get_objective_value()
                if optimal_val is None:
                    optimal_val = sol_obj_val
                if optimal_val is not None:
                    # sub-optimal solutions
                    if sol_obj_val != optimal_val:
                        continue
                opt_sol_count += 1

                # Translate solution to variables
                for cpm_var in self.user_vars:
                    solver_val = solution.get_value(self.solver_var(cpm_var))
                    if cpm_var.is_bool():
                        cpm_var._value = solver_val >= 0.5
                    else:
                        cpm_var._value = int(solver_val)

                # Translate objective
                if self.has_objective():
                    self.objective_value_ = sol_obj_val

                if display is not None:
                    if isinstance(display, Expression):
                        print(argval(display))
                    elif isinstance(display, list):
                        print(argvals(display))
                    else:
                        display()  # callback

        # Reset pool search mode to default
        self.cplex_model.context.cplex_parameters.mip.limits.populate = 1
        self.cplex_model.context.cplex_parameters.mip.pool.intensity = 0
        self.cplex_model.context.cplex_parameters.mip.pool.absgap = 1e-6  # Default value
        self.cplex_model.context.cplex_parameters.mip.pool.relgap = 1e-4  # Default value

        cplex_status = self.cplex_model.solve_details.status
        if cplex_status == "JobFailed": # something went wrong
            self.cpm_status.exitstatus = ExitStatus.ERROR
        elif "aborted" in cplex_status: # solve got prematurely aborted
            self.cpm_status.exitstatus = ExitStatus.NOT_RUN
        elif opt_sol_count: # found at least a single optimal solution
            if cplex_status == "populate solution limit exceeded": # reached solution limit
                # Sanity check
                if opt_sol_count == solution_limit: # reach set solution limit
                    self.cpm_status.exitstatus = ExitStatus.FEASIBLE
                else:
                    raise ValueError(f"cplex returned status {cplex_status}, but solution count {opt_sol_count} doesn't match set limit of {solution_limit}. Please report on GitHub.")
            elif "all reachable solutions enumerated" in cplex_status or cplex_status == "integer optimal solution": # found all solutions
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            elif cplex_status == "Unknown" or cplex_status == "time limit exceeded": # reached time limit
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            else:
                raise NotImplementedError(f"Translation of cplex status {cplex_status} to CPMpy status not implemented. Please report on GitHub")
        else:
            if cplex_status == "Unknown": # reached time limit
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            elif cplex_status == "integer infeasible": # Did not find any solution
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            else: 
                raise NotImplementedError(f"Translation of cplex status {cplex_status} to CPMpy status not implemented. Please report on GitHub")

        return opt_sol_count